import re

import numpy as np
import pytest
from mne import pick_info
from mne.datasets import testing
from mne.io import read_raw
from mne.preprocessing import ICA
from mne.utils import check_version

if check_version("mne", "1.6"):
    from mne._fiff.pick import _pick_data_channels, _picks_to_idx
else:
    from mne.io.pick import _pick_data_channels, _picks_to_idx

from tmseegpy.mne_icalabel.features import get_topomaps

directory = testing.data_path() / "MEG" / "sample"
raw = read_raw(directory / "sample_audvis_trunc_raw.fif", preload=False)
raw.pick_types(eeg=True, meg=True)
raw.load_data()
ica = ICA(n_components=5, method="picard")
ica.fit(raw)
ica_eeg = ICA(n_components=5, method="picard")
ica_eeg.fit(raw.pick_types(eeg=True))


@pytest.mark.filterwarnings("ignore:invalid value encountered in divide:RuntimeWarning")
@pytest.mark.parametrize("ica", [ica, ica_eeg])
def test_topomap_defaults(ica):
    """Test scalp topography array generation."""
    topomaps = get_topomaps(ica, picks=None)
    assert isinstance(topomaps, dict)
    for topomaps_ in topomaps.values():
        assert isinstance(topomaps_, np.ndarray)
        assert topomaps_.shape == (ica.n_components_, 64, 64)
    ch_picks = _pick_data_channels(ica.info, exclude=())
    ch_types = pick_info(ica.info, ch_picks).get_channel_types(unique=True)
    assert sorted(topomaps) == sorted(ch_types)


@pytest.mark.filterwarnings("ignore:invalid value encountered in divide:RuntimeWarning")
@pytest.mark.parametrize(
    ("picks", "res"),
    [(0, 32), ([0, 1, 2], 50), (slice(1, 3), 128), (np.array([1, 2]), 10)],
)
def test_topomap_arguments(picks, res):
    """Test arguments that influence the output shape."""
    topomaps = get_topomaps(ica, picks=picks, res=res)
    assert isinstance(topomaps, dict)
    ic_picks = _picks_to_idx(ica.n_components_, picks)
    for topomaps_ in topomaps.values():
        assert isinstance(topomaps_, np.ndarray)
        assert topomaps_.shape == (ic_picks.size, res, res)


@pytest.mark.filterwarnings("ignore:invalid value encountered in divide:RuntimeWarning")
def test_interpolation_arguments():
    """Test arguments that influence the interpolation."""
    # default
    topomaps1 = get_topomaps(ica)

    # image_interp
    topomaps2 = get_topomaps(ica, image_interp="linear")
    topomaps3 = get_topomaps(ica, image_interp="nearest")
    for topomaps in (topomaps2, topomaps3):
        assert sorted(topomaps1) == sorted(topomaps)
        for ch_type in topomaps1:
            assert not np.allclose(topomaps1[ch_type], topomaps[ch_type])
    del topomaps2, topomaps3, topomaps

    # border
    topomaps4 = get_topomaps(ica, border=5)
    assert sorted(topomaps1) == sorted(topomaps4)
    for ch_type in topomaps1:
        assert not np.allclose(topomaps1[ch_type], topomaps4[ch_type])
    del topomaps4

    # extrapolate
    topomaps5 = get_topomaps(ica, extrapolate="box")
    topomaps6 = get_topomaps(ica, extrapolate="head")
    topomaps7 = get_topomaps(ica, extrapolate="local")
    for topomaps in (topomaps5, topomaps6, topomaps7):
        assert sorted(topomaps1) == sorted(topomaps)
    del topomaps
    for ch_type in topomaps1:
        assert not np.allclose(topomaps1[ch_type], topomaps5[ch_type])
        if ch_type == "eeg":
            assert np.allclose(topomaps1[ch_type], topomaps6[ch_type])
        else:
            assert not np.allclose(topomaps1[ch_type], topomaps6[ch_type])
        if ch_type in ("grad", "mag"):
            assert np.allclose(topomaps1[ch_type], topomaps7[ch_type])
        else:
            assert not np.allclose(topomaps1[ch_type], topomaps7[ch_type])
    del topomaps5, topomaps6, topomaps7
    del topomaps1


def test_invalid_arguments():
    """Test invalid arguments."""
    with pytest.raises(TypeError, match="ica must be an instance of ICA"):
        get_topomaps(101)
    with pytest.raises(RuntimeError, match="The provided ICA instance was not fitted."):
        get_topomaps(ICA(n_components=5, method="picard"))

    with pytest.raises(
        TypeError,
        match=re.escape(
            "picks must be a list of int (indices) or list of str (names)."
        ),
    ):
        get_topomaps(ica, picks=101 + 101j)
    with pytest.raises(TypeError, match="Strings are not supported."):
        get_topomaps(ica, picks="101")
    with pytest.raises(IndexError, match="All picks must be < n_channels"):
        get_topomaps(ica, picks=6)

    with pytest.raises(TypeError, match="res must be an int"):
        get_topomaps(ica, res="101")
    with pytest.raises(TypeError, match="res must be an int"):
        get_topomaps(ica, res=True)
    with pytest.raises(ValueError, match="strictly positive integer. Provided '-101'"):
        get_topomaps(ica, res=-101)



from mne import BaseEpochs as BaseEpochs
from mne.io import BaseRaw as BaseRaw
from mne.preprocessing import ICA

from .config import ICALABEL_METHODS as ICALABEL_METHODS
from .iclabel._config import ICLABEL_NUMERICAL_TO_STRING as ICLABEL_NUMERICAL_TO_STRING
from .utils._checks import _validate_inst_and_ica as _validate_inst_and_ica

def label_components(inst: BaseRaw | BaseEpochs, ica: ICA, method: str):
    """Automatically label the ICA components with the selected method.

    Parameters
    ----------
    inst : Raw | Epochs
        The data instance used to fit the ICA instance.
    ica : ICA
        The fitted ICA instance.
    method : str
        The proposed method for labeling components. Must be one of:
        ``'iclabel'``.

    Returns
    -------
    component_dict : dict
        A dictionary with the following fields:

        - 'y_pred_proba' : array of shape (n_components,)
              Estimated predicted probability of the output class
              for each independent component.
        - 'labels': list of shape (n_components,)
              The corresponding string label of each class in 'y_pred'.

    Notes
    -----
    Please refer to the following function for additional information on each
    method:

    - ``'iclabel'``: `~mne_icalabel.iclabel.iclabel_label_components`
    """



from types import ModuleType as ModuleType

_INSTALL_MAPPING: dict[str, str]

def import_optional_dependency(
    name: str, extra: str = "", raise_error: bool = True
) -> ModuleType | None:
    """Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice message will be
    raised.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    raise_error : bool
        What to do when a dependency is not found.
        * True : Raise an ImportError.
        * False: Return None.

    Returns
    -------
    module : Module | None
        The imported module when found.
        None is returned when the package is not found and raise_error is False.
    """



import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
import os
import mne
from scipy import signal
from typing import Dict, Tuple, List, Optional

# Define key TEP components we want to identify
TEP_COMPONENTS = {
    'N15-P30': {'time': (15, 40), 'polarity': 'negative'},
    'N45': {'time': (40, 50), 'polarity': 'negative'},
    'P60': {'time': (55, 70), 'polarity': 'positive'},
    'N100': {'time': (85, 140), 'polarity': 'negative'},
    'P180': {'time': (150, 250), 'polarity': 'positive'}
}

def find_peak_in_window(times, gfp, t_min, t_max, polarity='positive', prominence=0.1):
    """Find peak in a specific time window."""
    # Get data in time window
    mask = (times >= t_min) & (times <= t_max)
    window_gfp = gfp[mask]
    window_times = times[mask]
    
    # Invert signal for negative peaks
    if polarity == 'negative':
        window_gfp = -window_gfp
    
    # Find peaks
    peaks, properties = signal.find_peaks(window_gfp, prominence=prominence)
    
    if len(peaks) > 0:
        # Get highest prominence peak
        best_peak = peaks[np.argmax(properties['prominences'])]
        peak_time = window_times[best_peak]
        peak_amplitude = window_gfp[best_peak] * (-1 if polarity == 'negative' else 1)
        return peak_time, peak_amplitude
    return None, None

def analyze_tep_components(evoked: mne.Evoked, prominence: float = 0.1):
    """Analyze TEP components using GFP and find peaks."""
    # Get data type
    ch_types = list(set(evoked.get_channel_types()))
    data_type = 'csd' if 'csd' in ch_types else 'eeg'
    
    # Calculate GFP
    times = evoked.times * 1000  # Convert to ms
    data = evoked.get_data()
    gfp = np.std(data, axis=0)
    
    # Find components
    components = {}
    for name, criteria in TEP_COMPONENTS.items():
        t_min, t_max = criteria['time']
        peak_time, peak_amplitude = find_peak_in_window(
            times, gfp, t_min, t_max, 
            criteria['polarity'], 
            prominence=prominence * np.max(gfp)
        )
        
        if peak_time is not None:
            components[name] = {
                'time': peak_time,
                'amplitude': peak_amplitude,
                'data_type': data_type
            }
    
    return components, gfp, times

def plot_tep_analysis(evoked: mne.Evoked, 
                     output_dir: str, 
                     session_name: str,
                     prominence: float = 0.1):
    """Create comprehensive TEP analysis plot with butterfly plot, GFP and topomaps."""
    components, gfp, times = analyze_tep_components(evoked, prominence)
    
    # Create figure with modified layout
    fig = plt.figure(figsize=(15, 10))
    
    # Create grid with small bottom area for topomaps
    gs = plt.GridSpec(2, 1, height_ratios=[3, 1], hspace=0.3)
    
    # Create subplot for main plots (butterfly and GFP)
    gs_main = gs[0].subgridspec(2, 1, height_ratios=[1.5, 1], hspace=0.3)
    
    # Plot butterfly plot with GFP
    ax_butterfly = fig.add_subplot(gs_main[0])
    evoked.plot(gfp=True, axes=ax_butterfly, show=False)
    ax_butterfly.set_title('TEP Butterfly Plot with GFP')
    
    # Plot GFP with detected components
    ax_gfp = fig.add_subplot(gs_main[1])
    ax_gfp.plot(times, gfp, 'b-', label='GFP')
    
    # Plot detected components
    colors = ['r', 'g', 'b', 'm', 'c']
    for (name, comp), color in zip(components.items(), colors):
        ax_gfp.plot(comp['time'], abs(comp['amplitude']), 'o', color=color, 
                   label=f"{name} ({comp['time']:.0f} ms)")
        ax_gfp.axvline(comp['time'], color=color, alpha=0.2)
    
    ax_gfp.set_xlabel('Time (ms)')
    ax_gfp.set_ylabel(f"GFP ({'µV/m²' if comp['data_type']=='csd' else 'µV'})")
    ax_gfp.set_title('Global Field Power with TEP Components')
    ax_gfp.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    ax_gfp.grid(True, alpha=0.3)
    ax_gfp.set_xlim(-100, 400)
    
    # Create subplot for topomaps
    gs_topos = gs[1].subgridspec(1, len(components), wspace=0.3)
    
    for idx, (name, comp) in enumerate(components.items()):
        ax = fig.add_subplot(gs_topos[idx])
        try:
            evoked.plot_topomap(times=comp['time']/1000.0,  # Convert back to seconds
                              ch_type=comp['data_type'],
                              axes=ax,
                              show=False,
                              time_format=f'{name}\n{comp["time"]:.0f} ms',
                              colorbar=False)  # Hide individual colorbars
        except Exception as e:
            ax.text(0.5, 0.5, f"Could not plot\n{name}", ha='center', va='center')
    
    # Add a single colorbar at the right of the topomaps
    cax = fig.add_axes([0.92, 0.11, 0.02, 0.15])  # [left, bottom, width, height]
    plt.colorbar(ax.images[-1], cax=cax)
    
    plt.suptitle(f'TEP Analysis - {session_name}', y=0.95)
    
    # Adjust the layout to prevent overlap
    plt.subplots_adjust(right=0.85)
    
    # Save figure
    fig.savefig(os.path.join(output_dir, f'{session_name}_tep_analysis.png'),
                dpi=300, bbox_inches='tight')
    plt.close(fig)
    
    return components

def generate_validation_summary(components: Dict, 
                              output_dir: str, 
                              session_name: str):
    """Generate a validation summary."""
    expected_times = {
        'N15-P30': (15, 40),
        'N45': (40, 50),
        'P60': (55, 70),
        'N100': (85, 140),
        'P180': (150, 250)
    }
    
    summary_path = os.path.join(output_dir, f'{session_name}_validation_summary.txt')
    with open(summary_path, 'w') as f:
        f.write("TEP Validation Summary\n")
        f.write("=" * 50 + "\n\n")
        
        for name, comp in components.items():
            expected_range = expected_times[name]
            is_valid = expected_range[0] <= comp['time'] <= expected_range[1]
            
            f.write(f"{name}:\n")
            f.write(f"  Latency: {comp['time']:.1f} ms ")
            f.write(f"({'valid' if is_valid else 'outside expected range'})\n")
            f.write(f"  Expected range: {expected_range[0]}-{expected_range[1]} ms\n\n")



# preproc_vis.py
import os
import matplotlib.pyplot as plt
import mne
from pathlib import Path
import numpy as np
import datetime

def create_step_directory(output_dir: str, session_name: str, step_name: str) -> str:
    """Create a directory for a specific preprocessing step's plots."""
    step_dir = Path(output_dir) / session_name / 'preprocessing_steps' / step_name
    step_dir.mkdir(parents=True, exist_ok=True)
    return str(step_dir)

def plot_raw_segments(raw, output_dir: str, session_name: str, step_name: str, duration: float = 5.0, overlap: float = 0.0):
    step_dir = create_step_directory(output_dir, session_name, step_name)
    
    ch_types = list(set(raw.get_channel_types()))
    data_type = 'csd' if 'csd' in ch_types else 'eeg'
    
    total_duration = raw.times[-1]
    step = duration - overlap
    start_times = np.arange(0, total_duration - duration, step)
    
    for i, start in enumerate(start_times):
        with mne.viz.use_browser_backend("matplotlib"):
            fig = raw.plot(duration=duration, start=start, show=False)
            plt.suptitle(f"{step_name} ({data_type.upper()}) - Segment {i+1} ({start:.1f}s - {start+duration:.1f}s)")
            fig.savefig(os.path.join(step_dir, f'{data_type}_segment_{i+1:03d}.png'), 
                       dpi=300, bbox_inches='tight')
            plt.close(fig)

def plot_epochs_grid(epochs, output_dir: str, session_name: str, step_name: str, epochs_per_plot: int = 10):
    step_dir = create_step_directory(output_dir, session_name, step_name)
    
    ch_types = list(set(epochs.get_channel_types()))
    data_type = 'csd' if 'csd' in ch_types else 'eeg'
    
    n_epochs = len(epochs)
    n_plots = (n_epochs + epochs_per_plot - 1) // epochs_per_plot
    
    for plot_idx in range(n_plots):
        start_idx = plot_idx * epochs_per_plot
        end_idx = min(start_idx + epochs_per_plot, n_epochs)
        
        with mne.viz.use_browser_backend("matplotlib"):
            fig = epochs[start_idx:end_idx].plot(show=False)
            plt.suptitle(f"{step_name} ({data_type.upper()}) - Epochs {start_idx+1}-{end_idx}")
            fig.savefig(os.path.join(step_dir, f'{data_type}_epochs_{plot_idx+1:03d}.png'), 
                       dpi=300, bbox_inches='tight')
            plt.close(fig)




import sys
if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

import numpy as np
import pytest
import torch
from scipy.io import loadmat

from tmseegpy.mne_icalabel.datasets import icalabel
from tmseegpy.mne_icalabel.iclabel.network.utils import _format_input
from tmseegpy.mne_icalabel.utils._tests import requires_module

dataset_path = icalabel.data_path() / "iclabel"

# Network weights
torch_iclabel_path = files("mne_icalabel.iclabel.network") / "assets" / "ICLabelNet.pt"
onnx_network_file = files("mne_icalabel.iclabel.network") / "assets" / "ICLabelNet.onnx"
matconvnet_iclabel_path = dataset_path / "network/netICL.mat"

# Network forward pass input/output
matconvnet_fw_input_path = dataset_path / "network/network_input.mat"
matconvnet_fw_output_path = dataset_path / "network/network_output.mat"

# Features (similar to network_input)
features_raw_path = dataset_path / "features/features-raw.mat"
features_epo_path = dataset_path / "features/features-epo.mat"

# Features formatted
features_formatted_raw_path = dataset_path / "features/features-formatted-raw.mat"
features_formatted_epo_path = dataset_path / "features/features-formatted-epo.mat"

# ICLabel output
iclabel_output_raw_path = dataset_path / "iclabel-output-raw.mat"
iclabel_output_epo_path = dataset_path / "iclabel-output-epo.mat"


def test_weights_pytorch():
    """Compare the weights of pytorch model and matconvnet model."""
    network_python = torch.load(torch_iclabel_path, weights_only=True)
    network_matlab = loadmat(matconvnet_iclabel_path)

    # load weights from matlab network
    weights_matlab = network_matlab["params"]["value"][0, :]
    # format weights from matlab network to torch convention
    for k, weight in enumerate(weights_matlab):
        if weight.ndim == 4:
            weights_matlab[k] = weight.transpose((3, 2, 0, 1))
        elif weight.ndim == 3:
            weights_matlab[k] = weight.transpose((2, 0, 1))

    network_python_layers = [
        layer for layer in network_python.keys() if "seq" not in layer
    ]
    network_matlab_layers = [elt[0] for elt in network_matlab["params"]["name"][0, :]]

    # match layer names torch -> matconvnet
    keyword_mapping = {
        "img": "image",
        "psds": "psdmed",
        "autocorr": "autocorr",
        "weight": "kernel",
        "bias": "bias",
    }

    for python_layer in network_python_layers:
        split = python_layer.split(".")
        if len(split) == 2:
            _, param = split
            matlab_layer = f"discriminator_conv_{keyword_mapping[param]}"
        elif len(split) == 3:
            feature, idx, param = split
            feature = keyword_mapping[feature.split("_")[0]]
            idx = int(idx[-1])
            param = keyword_mapping[param]
            matlab_layer = f"discriminator_{feature}_layer{idx}_conv_{param}"
        else:
            raise ValueError("Unexpected layer name.")

        # find matlab_layer idx
        idx = network_matlab_layers.index(matlab_layer)
        # compare layers weights
        assert np.allclose(network_python[python_layer], weights_matlab[idx])


@requires_module("torch")
def test_network_outputs_pytorch():
    """Compare that the ICLabel network in pytorch and matlab outputs the same values.

    Compare the outputs for a common set of features (input to the forward pass).

    Notes
    -----
    The forward pass has been run in matconvnet with the same input features.
    The corresponding MATLAB code can be found in 'data/network_output.txt'.
    """
    from mne_icalabel.iclabel.network.torch import ICLabelNet

    # load features to use for the forward pass
    features = loadmat(matconvnet_fw_input_path)["input"][0, :]
    # features is a (6, ) array with:
    # - elements in position [0, 2, 4] -> 1-element array with the var name
    # - elements in position [1, 3, 5] -> data arrays
    assert "in_image" == features[0][0]
    images = features[1]
    assert "in_psdmed" == features[2][0]
    psd = features[3]
    assert "in_autocorr" == features[4][0]
    autocorr = features[5]

    # reshape the features to fit torch format
    images = np.transpose(images, (3, 2, 0, 1))
    psd = np.transpose(psd, (3, 2, 0, 1))
    autocorr = np.transpose(autocorr, (3, 2, 0, 1))

    # convert to tensors
    images = torch.from_numpy(images).float()
    psd = torch.from_numpy(psd).float()
    autocorr = torch.from_numpy(autocorr).float()

    # run the forward pass on pytorch
    iclabel_net = ICLabelNet()
    iclabel_net.load_state_dict(torch.load(torch_iclabel_path, weights_only=True))
    torch_labels = iclabel_net(images, psd, autocorr)
    torch_labels = torch_labels.detach().numpy()  # (30, 7)

    # load the matconvnet output of the forward pass on those 3 feature arrays
    matlab_labels = loadmat(matconvnet_fw_output_path)["labels"]  # (30, 7)

    # Compare both outputs
    assert np.allclose(matlab_labels, torch_labels, atol=1e-7)


@requires_module("onnxruntime")
def test_network_outputs_onnx():
    """Compare that the ICLabel network in onnx and matlab outputs the same values.

    Compare the outputs for a common set of features (input to the forward pass).

    Notes
    -----
    The forward pass has been run in matconvnet with the same input features.
    The corresponding MATLAB code can be found in 'data/network_output.txt'.
    """
    import onnxruntime as ort

    # load features to use for the forward pass
    features = loadmat(matconvnet_fw_input_path)["input"][0, :]
    # features is a (6, ) array with:
    # - elements in position [0, 2, 4] -> 1-element array with the var name
    # - elements in position [1, 3, 5] -> data arrays
    assert "in_image" == features[0][0]
    images = features[1]
    assert "in_psdmed" == features[2][0]
    psd = features[3]
    assert "in_autocorr" == features[4][0]
    autocorr = features[5]

    # reshape the features to fit onnx format
    images = np.transpose(images, (3, 2, 0, 1)).astype(np.float32)
    psd = np.transpose(psd, (3, 2, 0, 1)).astype(np.float32)
    autocorr = np.transpose(autocorr, (3, 2, 0, 1)).astype(np.float32)

    # run the forward pass on onnxruntime
    ort_session = ort.InferenceSession(onnx_network_file)
    labels = ort_session.run(
        None,
        {"topo": images, "psds": psd, "autocorr": autocorr},
    )
    onnx_labels = labels[0]

    # load the matconvnet output of the forward pass on those 3 feature arrays
    matlab_labels = loadmat(matconvnet_fw_output_path)["labels"]  # (30, 7)

    # Compare both outputs
    assert np.allclose(matlab_labels, onnx_labels, atol=1e-7)


@pytest.mark.parametrize(
    ("eeglab_feature_file", "eeglab_feature_formatted_file"),
    [
        (features_raw_path, features_formatted_raw_path),
        (features_epo_path, features_formatted_epo_path),
    ],
)
def test_format_input(eeglab_feature_file, eeglab_feature_formatted_file):
    """Test formatting of input feature before feeding them to the network."""
    features_eeglab = loadmat(eeglab_feature_file)["features"]
    topo, psd, autocorr = _format_input(
        features_eeglab[0, 0], features_eeglab[0, 1], features_eeglab[0, 2]
    )

    features_formatted_eeglab = loadmat(eeglab_feature_formatted_file)["features"]
    topo_eeglab = features_formatted_eeglab[0, 0]
    psd_eeglab = features_formatted_eeglab[0, 1]
    autocorr_eeglab = features_formatted_eeglab[0, 2]

    assert np.allclose(topo, topo_eeglab)
    assert np.allclose(psd, psd_eeglab)
    assert np.allclose(autocorr, autocorr_eeglab)


@pytest.mark.parametrize(
    ("eeglab_feature_file", "eeglab_output_file"),
    [
        (features_raw_path, iclabel_output_raw_path),
        (features_epo_path, iclabel_output_epo_path),
    ],
)
@requires_module("onnxruntime")
def test_run_iclabel_onnx(eeglab_feature_file, eeglab_output_file):
    """Test that the network outputs the same values as MATLAB.

    Test the the network outputs the same values for the features in
    'features_raw_path' and 'features_epo_path' that contains the features extracted in
    EEGLAB. This set of feature is compared with the set of features retrieved in python
    in 'test_features.py:test_get_features'.
    """
    from mne_icalabel.iclabel.network.onnx import _run_iclabel

    features_eeglab = loadmat(eeglab_feature_file)["features"]
    # run the forward pass on onnx
    labels = _run_iclabel(
        features_eeglab[0, 0],
        features_eeglab[0, 1],
        features_eeglab[0, 2],
    )

    # load the labels from EEGLAB
    matlab_labels = loadmat(eeglab_output_file)["labels"]  # (30, 7)

    # Compare both outputs
    assert np.allclose(matlab_labels, labels, atol=1e-7)


@pytest.mark.parametrize(
    ("eeglab_feature_file", "eeglab_output_file"),
    [
        (features_raw_path, iclabel_output_raw_path),
        (features_epo_path, iclabel_output_epo_path),
    ],
)
@requires_module("torch")
def test_run_iclabel_pytorch(eeglab_feature_file, eeglab_output_file):
    """Test that the network outputs matches MATLAB.

    Test that the net work outputs the same values for the features in
    'features_raw_path' and 'features_epo_path' that contains the features extracted in
    EEGLAB. This set of feature is compared with the set of features retrieved in python
    in 'test_features.py:test_get_features'.
    """
    from mne_icalabel.iclabel.network.torch import _run_iclabel

    features_eeglab = loadmat(eeglab_feature_file)["features"]
    # run the forward pass on pytorch
    labels = _run_iclabel(
        features_eeglab[0, 0],
        features_eeglab[0, 1],
        features_eeglab[0, 2],
    )

    # load the labels from EEGLAB
    matlab_labels = loadmat(eeglab_output_file)["labels"]  # (30, 7)

    # Compare both outputs
    assert np.allclose(matlab_labels, labels, atol=1e-7)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (loader.py) is part of neurone_loader                             -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------
"""
Provides classes to load, represent and export data recorded with the Bittium NeurOne device.
"""

import os
import pandas as pd
import numpy as np
from operator import indexOf

from . import neurone as nr
from .lazy import Lazy, preloadable
from .mne_export import MneExportable
from .util import logger, doc_inherit


# noinspection PyAbstractClass
class BaseContainer(MneExportable):
    """
    A metaclass that provides properties for accessing data shared between all subclasses. I cannot be used itself
    as it is not implementing all required methods of its abstract superclass.
    """

    def __init__(self):
        self._dropped_channels = set()

    @property
    def sampling_rate(self):
        """
        :return: the sampling rate, read from the session protocol
        :rtype: int
        """
        return self._protocol['meta']['sampling_rate'] if hasattr(self, '_protocol') else None

    @property
    def subject_info(self):
        """
        :return: subject information, read from the session protocol. Possible keys are
                 `id`, `first_name`, `last_name`, `date_of_birth`
        :rtype: dict
        """
        return self._protocol['meta']['subject'] if hasattr(self, '_protocol') else None

    @property
    def time_start(self):
        """
        :return: recording start, read from the session protocol
        :rtype: datetime
        """
        return self._time_start if hasattr(self, '_time_start') else None

    @property
    def time_stop(self):
        """
        :return: recording end, read from the session protocol
        :rtype: datetime
        """
        return self._time_stop if hasattr(self, '_time_stop') else None

    def _protocol_channels(self):
        return self._protocol['channels'] if hasattr(self, '_protocol') else []

    def _drop_indexes(self):
        return sorted([indexOf(self._protocol_channels(), channel) for channel in self._dropped_channels], reverse=True)

    @Lazy
    def channels(self):
        """
        :return: ordered list of all channel names, read from the session protocol
        :rtype: list[str]
        """
        return [channel for channel in self._protocol_channels() if channel not in self._dropped_channels]

    def _has_data(self):
        private_attribute_name = getattr(type(self), 'data').private_name
        return hasattr(self, private_attribute_name)

    def _extend_droplist(self, channels_to_drop):
        self._dropped_channels |= set(channels_to_drop)
        if type(getattr(type(self), 'channels')) is Lazy:
            if hasattr(self, getattr(type(self), 'channels').private_name):
                self.channels = [channel for channel in self.channels if channel not in self._dropped_channels]

    def drop_channels(self, channels_to_drop):
        """
        Remove specified channels from loaded data. Dropped channels will be remembered and when data is cleared from
        memory and reloaded from disk the channels will get removed again. To get them back create a new object of this
        type to reload from disk.

        :param channels_to_drop: names of channels to drop
        :type channels_to_drop: list[str]
        """
        drop_set = (set(channels_to_drop) - self._dropped_channels).intersection(set(self.channels))
        not_dropping = set(channels_to_drop) - drop_set
        if len(not_dropping) > 0:
            logger.warning('Not dropping channels {channels} since they don\'t exist or have already been dropped'
                           .format(channels=', '.join(not_dropping)))
        logger.debug('Dropping channels {channels}'.format(channels=', '.join(drop_set)))
        self._extend_droplist(drop_set)


@preloadable
class Phase(BaseContainer):
    """
    Represents one recording phase of one NeurOne session in one NeurOne Recording

    :param path: path to the recording *session* folder
    :param phase: phase object from a session protocol
    :type path: str
    :type phase: dict
    """
    def __init__(self, path, phase, protocol=None):
        BaseContainer.__init__(self)
        self.path = path
        self.number = phase['number']
        if protocol is None:
            self._protocol = nr.read_neurone_protocol(self.path)
        else:
            self._protocol = protocol
        self._time_start = phase['time_start']
        self._time_stop = phase['time_stop']

    @Lazy
    def events(self):
        """
        :return: recorded events with Revision, Type, SourcePort, ChannelNumber, Code, StartSampleIndex,
                 StopSampleIndex, DescriptionLength, DescriptionOffset, DataLength, DataOffset, StartTime, StopTime
        :rtype: pandas.DataFrame
        """
        return pd.DataFrame(nr.read_neurone_events(self.path, self.number, self.sampling_rate)['events'])
    
    @property
    def event_codes(self):
        """
        :return: all event codes used in the data as int32 in an numpy.ndarray
        :rtype: numpy.ndarray
        """
        return np.unique(self.events['Code'].values) if 'Code' in self.events else []

    @Lazy
    def data(self):
        """
        :return: recorded data with shape (samples, channels) in µV
        :rtype: numpy.ndarray
        """
        data = nr.read_neurone_data(self.path, self.number, self._protocol) / 1000  # data is nanovolts
        return np.delete(data, self._drop_indexes(), axis=1)

    @property
    def n_samples(self):
        """
        :return: the number of channels, inferred from the binary recording's file size
        :rtype: int
        """
        return nr.read_neurone_data_info(self.path, self.number, self._protocol).n_samples

    @property
    def n_channels(self):
        """
        :return: the number of channels, read from the session protocol
        :rtype: int
        """
        return nr.read_neurone_data_info(self.path,
                                         self.number,
                                         self._protocol).n_channels - len(self._dropped_channels)

    def clear_data(self):
        """
        Remove loaded data from memory
        """
        del self.data

    # noinspection PyMissingOrEmptyDocstring
    @doc_inherit
    def drop_channels(self, channels_to_drop):
        if self._has_data():
            drop_indexes = sorted([indexOf(self.channels, channel) for channel in channels_to_drop
                                   if channel in self.channels],
                                  reverse=True)  # ignore channels that were dropped before
            # noinspection PyAttributeOutsideInit
            self.data = np.delete(self.data, drop_indexes, axis=1)
        BaseContainer.drop_channels(self, channels_to_drop)


@preloadable
class Session(BaseContainer):
    """
    Represents one session in one NeurOne Recording and contains all of the session's phases

    :param path: path to the recording *session* folder
    :type path: str
    """
    def __init__(self, path):
        BaseContainer.__init__(self)
        self.path = path
        self._protocol = nr.read_neurone_protocol(self.path)
        self._get_meta()

    def _get_meta(self):
        self._time_start = self._protocol['meta']['time_start']
        self._time_stop = self._protocol['meta']['time_stop']
        assert len(self._protocol['phases']) > 0, \
            "Session at {} has no phases".format(self.path)
        self.phases = [Phase(self.path, p, self._protocol) for p in self._protocol['phases']]

    @property
    def event_codes(self):
        """
        :return: all event codes used in the data as int32 in an :class:`numpy.ndarray`
        :rtype: numpy.ndarray
        """
        return np.unique(np.concatenate([phase.event_codes for phase in self.phases]))

    @Lazy
    def data(self):
        """
        .. warning:: Calling this replaces the data attribute of the contained phases with a view on the concatenated
             data to save memory. Keep this in mind when manipulating the contained sessions.

        :return: concatenated data of all phases with shape (samples, channels) in µV
        :rtype: numpy.ndarray
        """
        phases = sorted(self.phases, key=lambda phase: phase.number)
        new_array = None
        slices = []
        for p in phases:
            p.drop_channels(list(self._dropped_channels))
            if new_array is None:
                new_array = np.copy(p.data, order='C')
                slices.append((0, len(p.data)))
            else:
                old_length = len(new_array)
                shape = list(new_array.shape)
                shape[0] += len(p.data)
                # data is explicitly copied above and following slices are
                new_array.resize(shape, refcheck=False)
                # appended hence also copied, this should be fine
                new_array[-len(p.data):] = p.data # pylint: disable=unsupported-assignment-operation
                slices.append((old_length, old_length + len(p.data)))
            del p.data

        for index, p in enumerate(phases):
            start, stop = slices[index]
            p.data = new_array[start:stop]

        return new_array

    def clear_data(self):
        """
        Remove loaded data in all phases from memory
        """
        for p in self.phases:
            p.clear_data()
        del self.data

    @property
    def events(self):
        """
        :return: concatenated events of all phases with Revision, Type, SourcePort, ChannelNumber, Code,
                 StartSampleIndex, StopSampleIndex, DescriptionLength, DescriptionOffset, DataLength, DataOffset,
                 StartTime, StopTime
        :rtype: pandas.DataFrame
        """
        phases = sorted(self.phases, key=lambda p: p.number)
        all_events = [phases[0].events]
        current_samples = phases[0].n_samples
        for i in range(1, len(phases)):
            if len(phases[i].events) > 0:
                cur_events = phases[i].events.copy()
                cur_events['StartSampleIndex'] += current_samples
                cur_events['StopSampleIndex'] += current_samples
                cur_time = int(current_samples / self.sampling_rate)
                cur_events['StartTime'] += cur_time
                cur_events['StopTime'] += cur_time
                current_samples += phases[i].n_samples
                all_events.append(cur_events)
        return pd.concat(all_events)
        
    @property
    def n_samples(self):
        """
        :return: sum of the number of samples, inferred from the binary recording's file size, of all phases
        :rtype: int
        """
        return sum([p.n_samples for p in self.phases])

    @property
    def n_channels(self):
        """
        Returns the number of channels used in all phases and makes sure they're equal

        :return: the number of channels, read from the session protocol
        :rtype: int
        """
        assert len(set([p.n_channels for p in self.phases])) <= 1, \
            "The number of channels shouldn't change between phases"
        return self.phases[0].n_channels if len(self.phases) > 0 else 0

    # noinspection PyMissingOrEmptyDocstring
    @doc_inherit
    def drop_channels(self, channels_to_drop):
        if self._has_data():
            # noinspection PyAttributeOutsideInit
            drop_indexes = sorted([indexOf(self.channels, channel) for channel in channels_to_drop
                                   if channel in self.channels],
                                  reverse=True)  # ignore channels that were dropped before
            # noinspection PyAttributeOutsideInit
            self.data = np.delete(self.data, drop_indexes, axis=1)
            p_offset = 0
            for phase in self.phases:
                phase._extend_droplist(channels_to_drop)
                data_length = phase.data.shape[0]
                phase.data = self.data[p_offset:p_offset + data_length]
                p_offset += data_length
        else:
            for phase in self.phases:
                phase.drop_channels(channels_to_drop)
        BaseContainer.drop_channels(self, channels_to_drop)


@preloadable
class Recording(BaseContainer):
    """
    Represents one NeurOne Recording and contains all of the recording's sessions

    :param path: path to the recording *recording* folder
    :type path: str
    """
    def __init__(self, path):
        BaseContainer.__init__(self)
        self.path = path
        self._find_sessions()
        
    def _find_sessions(self):
        session_dirs = [os.path.join(self.path, dirname)
                        for dirname in os.listdir(self.path)
                        if os.path.isdir(os.path.join(self.path, dirname))
                        and 'Protocol.xml' in os.listdir(os.path.join(self.path, dirname))]
        assert len(session_dirs) > 0, "No sessions found in {}".format(self.path)
        self.sessions = list(sorted([Session(path) for path in session_dirs], key=lambda s: s.time_start))
        self._time_start = self.sessions[0].time_start if len(self.sessions) > 0 else None
        self._time_stop = self.sessions[-1].time_stop if len(self.sessions) > 0 else None

    @property
    def event_codes(self):
        """
        :return: all event codes used in the data as int32 in an numpy.ndarray
        :rtype: numpy.ndarray
        """
        return np.unique(np.concatenate([session.event_codes for session in self.sessions]))

    @Lazy
    def data(self):
        """
        :return: concatenated data of all phases of all sessions with shape (samples, channels) in µV
        :rtype: numpy.ndarray

        .. warning:: Calling this replaces the data attribute of the contained phases and sessions with a view on the
                     concatenated data to save memory. Keep this in mind when manipulating the contained sessions or
                     phases.
        """
        sessions = sorted(self.sessions, key=lambda x: x.time_start)
        new_array = None
        slices = []
        all_phase_slices = []
        for s in sessions:
            old_length = len(new_array) if new_array is not None else 0
            if s._has_data():
                new_length = old_length + len(s.data)
            else:
                new_length = old_length + s.n_samples
            del s.data

            phases = sorted(s.phases, key=lambda phase: phase.number)
            phase_slices = []
            for p in phases:
                p.drop_channels(list(self._dropped_channels))
                if new_array is None:
                    new_array = np.copy(p.data, order='C')
                    phase_slices.append((0, len(p.data)))
                else:
                    old_phase_length = len(new_array)
                    shape = list(new_array.shape)
                    shape[0] += len(p.data)
                    # data is explicitly copied above and following slices are
                    new_array.resize(shape, refcheck=False)
                    # appended hence also copied, this should be fine
                    new_array[-len(p.data):] = p.data  # pylint: disable=unsupported-assignment-operation
                    phase_slices.append((old_phase_length, old_phase_length + len(p.data)))
                del p.data

            all_phase_slices.append(phase_slices)
            slices.append((old_length, new_length))

        for s_index, s in enumerate(sessions):
            for p_index, p in enumerate(s.phases):
                start, stop = all_phase_slices[s_index][p_index]
                p.data = new_array[start:stop]
            start, stop = slices[s_index]
            s.data = new_array[start:stop]

        return new_array

    def clear_data(self):
        """
        Remove loaded data in all phases of all sessions from memory
        """
        for s in self.sessions:
            s.clear_data()
        del self.data

    @property
    def events(self):
        """
        :return: concatenated events of all phases of all sessions with Revision, Type, SourcePort, ChannelNumber, Code,
                 StartSampleIndex, StopSampleIndex, DescriptionLength, DescriptionOffset, DataLength, DataOffset,
                 StartTime, StopTime
        :rtype: pandas.DataFrame
        """
        sessions = sorted(self.sessions, key=lambda x: x.time_start)
        assert len(set([s.sampling_rate for s in sessions])) >= 1, \
            'Loading Sessions with different sampling rates is not supported at this time'
        sampling_rate = sessions[0].sampling_rate
        all_events = [sessions[0].events]
        current_samples = sessions[0].n_samples
        for i in range(1, len(sessions)):
            if len(sessions[i].events) > 0:
                cur_events = sessions[i].events.copy()
                cur_events['StartSampleIndex'] += current_samples
                cur_events['StopSampleIndex'] += current_samples
                cur_time = int(current_samples / sampling_rate)
                cur_events['StartTime'] += cur_time
                cur_events['StopTime'] += cur_time
                current_samples += sessions[i].n_samples
                all_events.append(cur_events)
        return pd.concat(all_events)

    @property
    def n_samples(self):
        """
        :return: sum of the number of samples, inferred from the binary recording's file size, of all phases of all
                 sessions
        :rtype: int
        """
        return sum([s.n_samples for s in self.sessions])

    @property
    def n_channels(self):
        """
        Returns the number of channels used in all phases and makes sure they're equal

        :return: the number of channels, read from the session protocol
        :rtype: int
        """
        assert len(set([s.n_channels for s in self.sessions])) <= 1, \
            "The number of channels shouldn't change between sessions"
        return self.sessions[0].n_channels if len(self.sessions) > 0 else 0

    @property
    def sampling_rate(self):
        """
        Returns the sampling rate used in all sessions and makes sure they're all equal

        :return: the sampling rate, read from the session protocols
        :rtype: int
        """
        assert len(set([s.sampling_rate for s in self.sessions])) <= 1, \
            "The sampling rate shouldn't change between sessions"
        return self.sessions[0].sampling_rate if len(self.sessions) > 0 else 0

    @property
    def channels(self):
        """
        Returns the channels used in all sessions and makes sure they're equal

        :return: ordered list of all channel names, read from the session protocols
        :rtype: list[str]
        """
        assert len(set([''.join(s.channels) for s in self.sessions])) <= 1, \
            "Channel names shouldn't change between sessions"
        return [channel for channel in self.sessions[0].channels
                if channel not in self._dropped_channels] if len(self.sessions) > 0 else 0

    # noinspection PyMissingOrEmptyDocstring
    @doc_inherit
    def drop_channels(self, channels_to_drop):
        if self._has_data():
            # noinspection PyAttributeOutsideInit
            drop_indexes = sorted([indexOf(self.channels, channel) for channel in channels_to_drop
                                   if channel in self.channels],
                                  reverse=True)  # ignore channels that were dropped before
            # noinspection PyAttributeOutsideInit
            self.data = np.delete(self.data, drop_indexes, axis=1)
            s_offset = 0
            p_offset = 0
            # update views and droplists
            for session in self.sessions:
                session._extend_droplist(channels_to_drop)
                data_length = session.data.shape[0]
                session.data = self.data[s_offset:s_offset + data_length]
                s_offset += data_length
                for phase in session.phases:
                    phase._extend_droplist(channels_to_drop)
                    data_length = phase.data.shape[0]
                    phase.data = self.data[p_offset:p_offset + data_length]
                    p_offset += data_length
        else:
            for session in self.sessions:
                session.drop_channels(channels_to_drop)
        BaseContainer.drop_channels(self, channels_to_drop)



from __future__ import annotations  # c.f. PEP 563, PEP 649

from importlib.resources import files
from typing import TYPE_CHECKING

import numpy as np
import onnxruntime as ort

from .utils import _format_input

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray


def _format_input_for_onnx(topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike):
    """Format the features to the correct shape and type for ONNX."""
    topo = np.transpose(topo, (3, 2, 0, 1)).astype(np.float32)
    psd = np.transpose(psd, (3, 2, 0, 1)).astype(np.float32)
    autocorr = np.transpose(autocorr, (3, 2, 0, 1)).astype(np.float32)

    return topo, psd, autocorr


def _run_iclabel(images: ArrayLike, psds: ArrayLike, autocorr: ArrayLike) -> NDArray:
    """Run ICLabel using onnx."""
    # load weights
    network_file = files("mne_icalabel.iclabel.network") / "assets" / "ICLabelNet.onnx"
    session = ort.InferenceSession(network_file)
    # format inputs
    topo, psds, autocorr = _format_input_for_onnx(
        *_format_input(images, psds, autocorr)
    )
    # run the forward pass
    labels = session.run(
        None,
        {"topo": topo, "psds": psds, "autocorr": autocorr},
    )
    return labels[0]



from numpy.typing import ArrayLike
from numpy.typing import NDArray as NDArray

def _format_input(
    topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike
) -> tuple[NDArray, NDArray, NDArray]:
    """Replicate the input formatting in EEGLAB -ICLabel.

    .. code-block:: matlab

       images = cat(4, images, -images, images(:, end:-1:1, :, :), ...
                    -images(:, end:-1:1, :, :));
       psds = repmat(psds, [1 1 1 4]);
       autocorrs = repmat(autocorrs, [1 1 1 4]);
    """



import pytest
from mne import create_info
from mne.io import RawArray
from mne.preprocessing import ICA

from tmseegpy.mne_icalabel.utils._checks import _validate_ica, _validate_inst_and_ica


def test_validate_inst_and_ica(rng):
    """Test that validate_inst_and_ica correctly raises."""
    # check types
    with pytest.raises(TypeError, match="inst must be an instance of Raw or Epochs"):
        _validate_inst_and_ica(101, ICA())

    raw = RawArray(rng.integers(1, 10, (3, 1000)), create_info(3, 100, "eeg"))
    with pytest.raises(TypeError, match="ica must be an instance of ICA"):
        _validate_inst_and_ica(raw, 101)

    # test unfitted
    ica = ICA(n_components=3, method="picard")
    with pytest.raises(RuntimeError, match="The provided ICA instance was not fitted."):
        _validate_inst_and_ica(raw, ica)

    # to avoid RuntimeWarning with fitting an unfiltered raw, let's fake the filter
    with raw.info._unlock():
        raw.info["highpass"] = 1.0
    ica.fit(raw)
    # test valid
    _validate_inst_and_ica(raw, ica)


def test_validate_ica(rng):
    """Test that _validate_ica correctly raises."""
    # check types
    with pytest.raises(TypeError, match="ica must be an instance of ICA"):
        _validate_ica(101)

    # test unfitted
    ica = ICA(n_components=3, method="picard")
    with pytest.raises(RuntimeError, match="The provided ICA instance was not fitted."):
        _validate_ica(ica)

    # to avoid RuntimeWarning with fitting an unfiltered raw, let's fake the filter
    raw = RawArray(rng.integers(1, 10, (3, 1000)), create_info(3, 100, "eeg"))
    with raw.info._unlock():
        raw.info["highpass"] = 1.0
    ica.fit(raw)
    # test valid
    _validate_ica(ica)



from tmseegpy.mne_icalabel.datasets.icalabel import data_path


def test_dataset(tmp_path):
    """Test that the MNE-ICAlabel testing dataset is available."""
    # download
    datapath = data_path(path=tmp_path)
    assert datapath.is_dir()
    assert (datapath / "iclabel").is_dir()

    # no download
    datapath = data_path(path=tmp_path)
    assert datapath.is_dir()
    assert (datapath / "iclabel").is_dir()



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

from .iclabel import iclabel_label_components

if TYPE_CHECKING:
    from typing import Callable, Optional

ICALABEL_METHODS: dict[str, Optional[Callable]] = {
    "iclabel": iclabel_label_components,
    "manual": None,
}

# map labels to the equivalent str format in MNE
ICA_LABELS_TO_MNE: dict[str, str] = {
    "Brain": "brain",
    "Muscle": "muscle",
    "Eye": "eog",
    "Heart": "ecg",
    "Line Noise": "line_noise",
    "Channel Noise": "ch_noise",
    "Other": "other",
}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (util.py) is part of neurone_loader                               -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------
"""
Provides internal utility functions
"""

import logging
import sys
from functools import wraps


def get_logger():
    """
    Get the default logger for this module and set default settings.

    :return: the logger
    :rtype: logging.Logger
    """
    lgr = logging.getLogger('neurone_loader')
    lgr.setLevel(logging.INFO)
    fmt = logging.Formatter(fmt='%(asctime)s [%(name)s.%(module)s]:%(levelname)s: %(message)s', datefmt='%I:%M:%S')
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.NOTSET)
    stdout_handler.setFormatter(fmt)
    lgr.addHandler(stdout_handler)
    return lgr


logger = get_logger()


class DocInherit(object):
    """
    Docstring inheriting method descriptor

    The class itself is also used as a decorator
    """

    def __init__(self, method):
        self.method = method
        self.name = method.__name__

    def __get__(self, obj, cls):
        if obj:
            return self._get_with_inst(obj, cls)
        else:
            return self._get_no_inst(cls)

    def _get_with_inst(self, obj, cls):
        overridden = getattr(super(cls, obj), self.name, None)

        @wraps(self.method, assigned=('__name__', '__module__'))
        def _func(*args, **kwargs):
            return self.method(obj, *args, **kwargs)

        return self._use_parent_doc(_func, overridden)

    def _get_no_inst(self, cls):
        overridden = next((getattr(parent, self.name, None) for parent in cls.__mro__[1:]), None)

        @wraps(self.method, assigned=('__name__', '__module__'))
        def _func(*args, **kwargs):
            return self.method(*args, **kwargs)

        return self._use_parent_doc(_func, overridden)

    def _use_parent_doc(self, func, source):
        if source is None:
            raise NameError('Can\'t find {name} in parents'.format(name=self.name))
        func.__doc__ = source.__doc__
        return func


doc_inherit = DocInherit



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mne.preprocessing import ICA


def label_ica_components(inst, ica: ICA, show: bool = True, block: bool = False):
    """Launch the IC labelling GUI.

    Parameters
    ----------
    inst : Raw | Epochs
        `~mne.io.Raw` or `~mne.Epochs` instance used to fit the `~mne.preprocessing.ICA`
        decomposition.
    ica : ICA
        The ICA object fitted on ``inst``.
    show : bool
        Show the GUI if True.
    block : bool
        Whether to halt program execution until the figure is closed.

    Returns
    -------
    gui : instance of ICAComponentLabeler
        The graphical user interface (GUI) window.
    """
    from mne.viz.backends._utils import _init_mne_qtapp, _qt_app_exec

    try:
        from ._label_components import ICAComponentLabeler
    except ImportError as e:
        raise ImportError(f"{e}. Users must install Qt bindings on their own.")

    # get application
    app = _init_mne_qtapp()
    gui = ICAComponentLabeler(inst=inst, ica=ica, show=show)
    if block:
        _qt_app_exec(app)
    return gui



from numpy.typing import ArrayLike, NDArray

def run_iclabel(
    images: ArrayLike,
    psds: ArrayLike,
    autocorr: ArrayLike,
    backend: str | None = "pytorch",
) -> NDArray:
    """Run the ICLabel network on the provided set of features.

    The features are un-formatted and are as-returned by
    `~mne_icalabel.iclabel.get_iclabel_features`. For more information,
    see :footcite:t:`PionTonachini2019`.

    Parameters
    ----------
    images : array of shape (n_components, 1, 32, 32)
        The topoplot images.
    psds : array of shape (n_components, 1, 1, 100)
        The power spectral density features.
    autocorr : array of shape (n_components, 1, 1, 100)
        The autocorrelation features.
    backend : None | ``torch`` | ``onnx``
        Backend to use to run ICLabel. If None, returns the first available backend in
        the order ``torch``, ``onnx``.

    Returns
    -------
    labels : array of shape (n_components, n_classes)
        The predicted numerical probability values for all labels in ICLabel output.
        Columns are ordered with ``'Brain'``, ``'Muscle'``, ``'Eye'``,
        ``'Heart'``, ``'Line Noise'``, ``'Channel Noise'``, and ``'Other'``.

    References
    ----------
    .. footbibliography::
    """



from __future__ import annotations

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

from typing import TYPE_CHECKING

import numpy as np
import torch
import torch.nn as nn

from .utils import _format_input

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray


class _ICLabelNetImg(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            in_channels=1,
            out_channels=128,
            kernel_size=(4, 4),
            padding=1,
            stride=(2, 2),
        )
        self.relu1 = nn.LeakyReLU(negative_slope=0.2)
        self.conv2 = nn.Conv2d(
            in_channels=128,
            out_channels=256,
            kernel_size=(4, 4),
            padding=1,
            stride=(2, 2),
        )
        self.relu2 = nn.LeakyReLU(negative_slope=0.2)
        self.conv3 = nn.Conv2d(
            in_channels=256,
            out_channels=512,
            kernel_size=(4, 4),
            padding=1,
            stride=(2, 2),
        )
        self.relu3 = nn.LeakyReLU(negative_slope=0.2)
        self.sequential = nn.Sequential(
            self.conv1, self.relu1, self.conv2, self.relu2, self.conv3, self.relu3
        )

    def forward(self, x):  # noqa: D102
        return self.sequential(x)


class _ICLabelNetPSDS(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            in_channels=1,
            out_channels=128,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu1 = nn.LeakyReLU(negative_slope=0.2)
        self.conv2 = nn.Conv2d(
            in_channels=128,
            out_channels=256,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu2 = nn.LeakyReLU(negative_slope=0.2)
        self.conv3 = nn.Conv2d(
            in_channels=256,
            out_channels=1,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu3 = nn.LeakyReLU(negative_slope=0.2)
        self.sequential = nn.Sequential(
            self.conv1, self.relu1, self.conv2, self.relu2, self.conv3, self.relu3
        )

    def forward(self, x):  # noqa: D102
        return self.sequential(x)


class _ICLabelNetAutocorr(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            in_channels=1,
            out_channels=128,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu1 = nn.LeakyReLU(negative_slope=0.2)
        self.conv2 = nn.Conv2d(
            in_channels=128,
            out_channels=256,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu2 = nn.LeakyReLU(negative_slope=0.2)
        self.conv3 = nn.Conv2d(
            in_channels=256,
            out_channels=1,
            kernel_size=(1, 3),
            padding=(0, 1),
            stride=(1, 1),
        )
        self.relu3 = nn.LeakyReLU(negative_slope=0.2)
        self.sequential = nn.Sequential(
            self.conv1, self.relu1, self.conv2, self.relu2, self.conv3, self.relu3
        )

    def forward(self, x):  # noqa: D102
        return self.sequential(x)


class ICLabelNet(nn.Module):
    """The ICLabel neural network."""

    def __init__(self):
        super().__init__()

        self.img_conv = _ICLabelNetImg()
        self.psds_conv = _ICLabelNetPSDS()
        self.autocorr_conv = _ICLabelNetAutocorr()

        self.conv = nn.Conv2d(
            in_channels=712,
            out_channels=7,
            kernel_size=(4, 4),
            padding=0,
            stride=(1, 1),
        )
        self.softmax = nn.Softmax(dim=1)

        self.seq = nn.Sequential(self.conv, self.softmax)

    @staticmethod
    def reshape_fortran(x: torch.Tensor, shape) -> torch.Tensor:  # noqa: D102
        x = x.permute(*reversed(range(len(x.shape))))
        return x.reshape(*reversed(shape)).permute(*reversed(range(len(shape))))

    def reshape_concat(self, tensor: torch.Tensor) -> torch.Tensor:  # noqa: D102
        tensor = self.reshape_fortran(tensor, [-1, 1, 1, 100])
        tensor = torch.cat([tensor, tensor, tensor, tensor], 1)
        tensor = torch.cat([tensor, tensor, tensor, tensor], 2)
        tensor = tensor.permute((0, 3, 1, 2))
        return tensor

    def forward(  # noqa: D102
        self, images: torch.Tensor, psds: torch.Tensor, autocorr: torch.Tensor
    ) -> torch.Tensor:
        out_img = self.img_conv(images)
        out_psds = self.psds_conv(psds)
        out_autocorr = self.autocorr_conv(autocorr)

        # PSDS reshape, concat, permute
        psds_perm = self.reshape_concat(out_psds)

        # Autocorr reshape, concat, permute
        autocorr_perm = self.reshape_concat(out_autocorr)

        concat = torch.cat([out_img, psds_perm, autocorr_perm], 1)

        labels = self.seq(concat)

        labels = labels.squeeze()
        labels = self.reshape_fortran(labels.permute(1, 0), [-1, 4])
        labels = torch.mean(labels, 1)
        labels = self.reshape_fortran(labels, [7, -1])
        labels = labels.permute(1, 0)

        return labels


def _format_input_for_torch(topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike):
    """Format the features to the correct shape and type for pytorch."""
    topo = np.transpose(topo, (3, 2, 0, 1))
    psd = np.transpose(psd, (3, 2, 0, 1))
    autocorr = np.transpose(autocorr, (3, 2, 0, 1))

    topo = torch.from_numpy(topo).float()
    psd = torch.from_numpy(psd).float()
    autocorr = torch.from_numpy(autocorr).float()

    return topo, psd, autocorr


def _run_iclabel(images: ArrayLike, psds: ArrayLike, autocorr: ArrayLike) -> NDArray:
    """Run ICLabel using onnx."""
    # load weights
    network_file = files("tmseegpy.mne_icalabel.iclabel.network") / "assets" / "ICLabelNet.pt"
    iclabel_net = ICLabelNet()
    iclabel_net.load_state_dict(torch.load(network_file, weights_only=True))
    # format inputs and run forward pass
    labels = iclabel_net(
        *_format_input_for_torch(*_format_input(images, psds, autocorr))
    )
    return labels.detach().numpy()



from pathlib import Path as Path

from mne.preprocessing import ICA

from ..config import ICA_LABELS_TO_MNE as ICA_LABELS_TO_MNE
from ..iclabel._config import ICLABEL_STRING_TO_NUMERICAL as ICLABEL_STRING_TO_NUMERICAL
from ..utils._imports import import_optional_dependency as import_optional_dependency

def write_components_tsv(ica: ICA, fname: str | Path):
    """Write channels tsv file for ICA components.

    Will create an accompanying JSON sidecar to explain the
    additional columns created in the channels tsv file for the
    ICA component labeling.

    Parameters
    ----------
    ica : ICA
        An instance of the fitted ICA.
    fname : str | Path
        The output filename.

    Notes
    -----
    Components are stored in a ``.tsv`` file essentially in the same manner as
    ``channels.tsv`` files for BIDS-EEG data. For more information, see the
    `BIDS specification <https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/03-electroencephalography.html#channels-description-_channelstsv>`_
    for EEG channels metadata.

    Storage of ICA annotations as a ``.tsv`` file is currently experimental in the
    context of BIDS-EEG Derivatives. The API and functionality is subject to change
    as the community converges on the specification of BIDS-Derivatives.
    """

def mark_component(
    component: int,
    fname: str | Path,
    method: str,
    label: str,
    author: str,
    strict_label: bool = True,
):
    """Mark a component with a label.

    Parameters
    ----------
    component : int
        The component to mark.
    fname : Union[str, Path]
        The filename for the BIDS filepath.
    method : str
        The method to use. Must be 'manual', or one of ['iclabel'].
    label : str
        The label of the ICA component. Must be one of ['brain',
        'muscle artifact', 'eye blink', 'heart beat', 'line noise',
        'channel noise', 'other'].
    author : str
        The annotating author.
    strict_label : bool
        Whether to raise an error if ``label`` is not an accepted value.
        Default is True.

    Notes
    -----
    Storage of ICA annotations as a ``.tsv`` file is currently experimental in the
    context of BIDS-EEG Derivatives. The API and functionality is subject to change
    as the community converges on the specification of BIDS-Derivatives.
    """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (lazy.py) is part of neurone_loader                               -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------
"""
Provides the `Lazy` decorator to construct properties that are evaluated only once and the
`preloadable` decorator to enable optional preloading of all lazy properties on initialization.
"""

from functools import update_wrapper
from .util import logger


# noinspection PyMethodOverriding
class Lazy(property):
    """
    Return a lazy property attribute.

    This decorator can be used exactly like the :py:class:`property` function to turn a function into an attribute, the
    difference being the following: A function decorated with :py:class:`property` is evaluated every time the attribute
    is accessed. A function decorated with :py:class:`.lazy.Lazy` is only evaluated once and the result is stored as a
    private attribute. Subsequently the private attribute is returned when the property constructed with
    :py:class:`.lazy.Lazy` is accessed. The lazy property can also be set manually or deleted, just like every other
    attribute. When the lazy attribute is deleted and then accessed again, the property function is called again and
    the result stored as a private attribute.

    :Example:

    >>> class Test:
    >>>     @Lazy
    >>>     def lazy_attribute(self):
    >>>         print('lazy function called')
    >>>         return 'lazy return'
    >>>
    >>>     @property
    >>>     def property_attribute(self):
    >>>         print('property function called')
    >>>         return 'property return'
    >>>
    >>> test_object = Test()
    >>> print(test_object.property_attribute)
    property function called
    property return
    >>> print(test_object.property_attribute) # A property function is evaluated on every call
    property function called
    property return
    >>> print(test_object.lazy_attribute) # The lazy function is evaluated on first call
    lazy function called
    lazy return
    >>> print(test_object.lazy_attribute) # but not on subsequent calls
    lazy return
    >>> del test_object.lazy_attribute    # When deleted the attribute is reset and the
    >>> print(test_object.lazy_attribute) # function is evaluated again on next call
    lazy function called
    lazy return

    .. seealso:: Decorate your class with the :py:class:`..lazy.preloadable` attribute to enable optional preloading of
                 all lazy attributes on initialization.
    """
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.private_name = "_{}".format(fget.__name__)
        doc = doc or fget.__doc__
        property.__init__(self, fget=fget, fset=fset, fdel=fdel, doc=doc)

        # noinspection PyTypeChecker
        update_wrapper(self, fget)

        if type(self.__doc__) is str:
            self.__doc__ = """
            .. note:: This property is a lazy property. For details see :py:class:`.lazy.Lazy`
            
            {old_doc}
            """.format(old_doc=self.__doc__)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if hasattr(instance, self.private_name):
            result = getattr(instance, self.private_name)
        else:
            logger.debug('(Lazy) loading {}.{}'.format(owner.__name__, self.fget.__name__))
            # noinspection PyArgumentList
            result = self.fget(instance)
            setattr(instance, self.private_name, result)
        return result

    def __set__(self, instance, value):
        if self.fset is None:
            setattr(instance, self.private_name, value)
        else:
            # noinspection PyArgumentList
            self.fset(instance, value)

    def __delete__(self, instance):
        if self.fdel is None:
            if hasattr(instance, self.private_name):
                delattr(instance, self.private_name)
        else:
            # noinspection PyArgumentList
            self.fdel(instance)


def preloadable(cls):
    """
    Use this as a decorator for a class that contains properties constructed with :py:class:`.lazy.Lazy`. A class
    decorated like this can be initialized with ``preload=True`` to call every lazy property once and store it's return
    value.
    Optionally the :py:class:`preload` function can be used to do the same. It can also be used to reload all lazy
    properties without deleting them first.

    :Example:

    >>> @preloadable
    >>> class Test:
    >>>     @Lazy
    >>>     def lazy_attribute(self):
    >>>         print('lazy function called')
    >>>         return 'lazy return'
    >>>
    >>> test_object = Test(preload=True) # The lazy property is evaluated on initialization
    lazy function called
    >>> print(test_object.lazy_attribute) # The stored attribute is returned
    lazy return
    >>> del test_object.lazy_attribute    # When deleted the attribute is reset and the
    >>> print(test_object.lazy_attribute) # function is evaluated again on next call
    lazy function called
    lazy return
    >>> test_object.preload() # All properties are reloaded even though already stored
    lazy function called
    """

    def _preload(self):
        """
        Use this function to call all properties constructed with :py:class:`.lazy.Lazy`. It can also be used to
        reload all lazy properties without deleting them first.

        :Example:

        >>> @preloadable
        >>> class Test:
        >>>     @Lazy
        >>>     def lazy_attribute(self):
        >>>         print('lazy function called')
        >>>         return 'lazy return'
        >>>
        >>> test_object = Test(preload=False) # The lazy property is not evaluated on initialization
        >>> test_object.preload()
        lazy function called
        >>> print(test_object.lazy_attribute) # The stored attribute is returned
        lazy return
        >>> test_object.preload() # All properties are reloaded even though already stored
        lazy function called
        """
        obj_type = type(self)

        def _try_preload_child(obj):
            if hasattr(obj, 'preload'):
                preload_function = getattr(obj, 'preload')
                if callable(preload_function):
                    obj.preload()

        for attr in [attr for attr in dir(self) if not attr.startswith('__')]:
            possible_prop = getattr(obj_type, attr, None)
            if isinstance(possible_prop, Lazy):
                if not hasattr(self, possible_prop.private_name):
                    logger.debug('Preloading property {} of {}'.format(attr, self))
                    getattr(self, attr)

            attr_obj = getattr(self, attr)
            try:
                _ = (e for e in attr_obj)  # Iterable
                for x in attr_obj:
                    _try_preload_child(x)
            except TypeError:
                # not iterable
                _try_preload_child(attr_obj)

    original_init = cls.__init__
    cls.preload = _preload

    def _new_init(self, *args, **kwargs):
        if 'preload' in kwargs:
            preload_enabled = kwargs['preload']
            del kwargs['preload']
        else:
            preload_enabled = False

        original_init(self, *args, **kwargs)

        if preload_enabled:
            self.preload()

    cls.__init__ = _new_init

    return cls



from typing import IO, Callable

from packaging.requirements import Requirement

def sys_info(fid: IO | None = None, developer: bool = False):
    """Print the system information for debugging.

    Parameters
    ----------
    fid : file-like | None
        The file to write to, passed to :func:`print`. Can be None to use
        :data:`sys.stdout`.
    developer : bool
        If True, display information about optional dependencies.
    """

def _list_dependencies_info(
    out: Callable, ljust: int, package: str, dependencies: list[Requirement]
) -> None:
    """List dependencies names and versions."""

def _get_gpu_info() -> tuple[str | None, str | None]:
    """Get the GPU information."""



from mne.io import BaseRaw
from numpy.typing import ArrayLike, NDArray

def _mne_to_eeglab_locs(
    raw: BaseRaw, picks: list[str]
) -> tuple[NDArray[float], NDArray[float]]:
    """Obtain EEGLab-like spherical coordinate from EEG channel positions.

    TODO: @JACOB:
    - Where is (0,0,0) defined in MNE vs EEGLab?
    - some text description of how the sphere coordinates differ between MNE
    and EEGLab.

    Parameters
    ----------
    raw : mne.io.BaseRaw
        Instance of raw object with a `mne.montage.DigMontage` set with
        ``n_channels`` channel positions.
    picks : list of str
        List of channel names to include.

    Returns
    -------
    Rd : np.array of shape (1, n_channels)
        Angle in spherical coordinates of each EEG channel.
    Th : np.array of shape (1, n_channels)
        Degree in spherical coordinates of each EEG channel.
    """

def _pol2cart(
    theta: NDArray[float], rho: NDArray[float]
) -> tuple[NDArray[float], NDArray[float]]:
    """Convert polar coordinates to cartesian coordinates.

    Parameters
    ----------
    theta : array
        angle
    rho : array
        magnitude
    """

def _next_power_of_2(x) -> int:
    """Equivalent to 2^nextpow2 in MATLAB."""

def _gdatav4(
    x: ArrayLike, y: ArrayLike, v: ArrayLike, xq: ArrayLike, yq: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    """GDATAV4 MATLAB 4 GRIDDATA interpolation.

    Parameters
    ----------
    x : array
        x-coordinates
    y : array
        y-coordinates
    v : array
        values
    xq : array
        x-grid
    yq : array
        y-grid

    Returns
    -------
    xq : array
    yq : array
    vq : array

    Reference
    ---------
    David T. Sandwell, Biharmonic spline interpolation of GEOS-3 and SEASAT
    altimeter data, Geophysical Research Letters, 2, 139-142, 1987.

    Describes interpolation using value of gradient of value in any dimension.
    """

def _mergepoints2D(
    x: ArrayLike, y: ArrayLike, v: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    """Averages values for points that are close to each other.

    Parameters
    ----------
    x : array
        x-coordinates
    y : array
        y-coordinates
    v : array
        values

    Returns
    -------
    x : array
    y : array
    v : array
    """

def _mergesimpts(
    data: ArrayLike, tols: list[ArrayLike], mode: str = "average"
) -> ArrayLike:
    """Merge similar points."""



from . import annotation as annotation
from . import config as config
from . import datasets as datasets
from . import features as features
from . import gui as gui
from . import iclabel as iclabel
from . import utils as utils
from ._version import __version__ as __version__
from .label_components import label_components as label_components
from .utils.config import sys_info as sys_info



# Standard scientific libraries
import numpy as np
from scipy import stats, signal
from scipy.interpolate import interp1d
from scipy.stats import zscore
import matplotlib.pyplot as plt
import seaborn as sns
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from scipy.spatial.distance import pdist, squareform
from scipy import optimize
import mne
from mne.preprocessing import compute_current_source_density
from mne.io.constants import FIFF

from typing import Optional, Union, Dict, List, Tuple, Any
import warnings
import os

# MNE imports
import mne
from mne.minimum_norm import (make_inverse_operator, 
                            apply_inverse,
                            write_inverse_operator)
from mne.preprocessing import compute_proj_ecg, compute_proj_eog
from mne import (compute_raw_covariance,
                read_source_spaces,
                setup_source_space,
                make_bem_model,
                make_bem_solution,
                make_forward_solution,
                read_trans,
                read_bem_solution)
from mne.viz import plot_alignment, plot_bem

# Required for ICA and component labeling
from .mne_icalabel import label_components
from mne.preprocessing import ICA

# Required for FASTER bad channel/epoch detection 
from mne_faster import find_bad_channels, find_bad_epochs, find_bad_channels_in_epochs

# Required for artifact cleaning (if using TMSArtifactCleaner)
from sklearn.preprocessing import StandardScaler
import tensorly as tl
from tensorly.decomposition import parafac, non_negative_parafac, tucker
from tqdm import tqdm


## Custom TMS-artefact cleaner
from .clean import TMSArtifactCleaner


class TMSEEGPreprocessor:
    """
    A class for preprocessing TMS-EEG data.
    
    This class implements a preprocessing pipeline for TMS-EEG data,
    including artifact removal, filtering, and data quality checks.
    
    Parameters
    ----------
    raw : mne.io.Raw
        The raw TMS-EEG data
    montage : str or mne.channels.montage.DigMontage, optional
        The EEG montage to use (default is 'standard_1020')
    ds_sfreq : float, optional
        The desired sampling frequency for resampling (default is 1000 Hz)
        
    Attributes
    ----------
    raw : mne.io.Raw
        The raw TMS-EEG data
    epochs : mne.Epochs
        The epoched TMS-EEG data
    montage : mne.channels.montage.DigMontage
        The EEG montage
    """
    
    def __init__(self, 
                raw: mne.io.Raw,
                montage: Union[str, mne.channels.montage.DigMontage] = 'standard_1020',
                ds_sfreq: float = 1000):
        
        self.raw = raw.copy()
        self.epochs = None
        self.evoked = None
        self.ds_sfreq = ds_sfreq

        self.processing_stage = {
            'initial_removal': False,
            'first_interpolation': False,
            'artifact_cleaning': False,
            'extended_removal': False,
            'final_interpolation': False
        }
        
        # Remove unused EMG channels if present
        for ch in self.raw.info['ch_names']:
            if ch.startswith('EMG'):
                self.raw.drop_channels(ch)
            elif ch.startswith('31'):
                self.raw.drop_channels(ch)
            elif ch.startswith('32'):
                self.raw.drop_channels(ch)
        
        # Channel name standardization
        ch_names = self.raw.ch_names
        rename_dict = {}
        for ch in ch_names:
            # Common naming variations
            if ch in ['31', '32']:
                continue  # Skip non-EEG channels
            if ch.upper() == 'FP1':
                rename_dict[ch] = 'Fp1'
            elif ch.upper() == 'FP2':
                rename_dict[ch] = 'Fp2'
            elif ch.upper() in ['FPZ', 'FPOZ']:
                rename_dict[ch] = 'Fpz'
            elif ch.upper() == 'POZ':
                rename_dict[ch] = 'POz'
            elif ch.upper() == 'PZ':
                rename_dict[ch] = 'Pz'
            elif ch.upper() == 'FCZ':
                rename_dict[ch] = 'FCz'
            elif ch.upper() == 'CPZ':
                rename_dict[ch] = 'CPz'
            elif ch.upper() == 'FZ':
                rename_dict[ch] = 'Fz'
            elif ch.upper() == 'CZ':
                rename_dict[ch] = 'Cz'
            elif ch.upper() == 'OZ':
                rename_dict[ch] = 'Oz'
        
        if rename_dict:
            print("Renaming channels to match standard nomenclature:")
            for old, new in rename_dict.items():
                print(f"  {old} -> {new}")
            self.raw.rename_channels(rename_dict)
        
        # Set montage with error handling
        if isinstance(montage, str):
            try:
                self.montage = mne.channels.make_standard_montage(montage)
            except ValueError as e:
                print(f"Warning: Could not create montage '{montage}': {str(e)}")
                print("Falling back to standard_1020 montage")
                self.montage = mne.channels.make_standard_montage('standard_1020')
        else:
            self.montage = montage
        
        try:
            # First try to set montage normally
            self.raw.set_montage(self.montage)
        except ValueError as e:
            print(f"\nWarning: Could not set montage directly: {str(e)}")
            
            # Get the channel types
            ch_types = {ch: self.raw.get_channel_types(picks=ch)[0] for ch in self.raw.ch_names}
            
            # Identify non-EEG channels
            non_eeg = [ch for ch, type_ in ch_types.items() if type_ not in ['eeg', 'unknown']]
            if non_eeg:
                print(f"\nFound non-EEG channels: {non_eeg}")
                print("Setting their types explicitly...")
                for ch in non_eeg:
                    self.raw.set_channel_types({ch: 'misc'})
            
            # Try setting montage again with on_missing='warn'
            try:
                self.raw.set_montage(self.montage, on_missing='warn')
                print("\nMontage set successfully with warnings for missing channels")
            except Exception as e2:
                print(f"\nWarning: Could not set montage even with warnings: {str(e2)}")
                print("Continuing without montage. Some functionality may be limited.")
        
        self.events = None
        self.event_id = None

        # Initialize attributes that will be set later
        self.stc = None
        self.forward = None
        self.inverse_operator = None
        self.source_space = None
        self.bem_solution = None
        self.noise_cov = None

        self.preproc_stats = {
            'n_orig_events': 0,
            'n_final_events': 0,
            'bad_channels': [],
            'n_bad_epochs': 0,
            'muscle_components': [],
            'excluded_ica_components': [],
            'original_sfreq': 0,
            'interpolated_times': [],
        }
        
        
        
    def create_epochs(self, 
                    tmin: float = -0.5, 
                    tmax: float = 1,
                    baseline: Optional[Tuple[float, float]] = None,
                    amplitude_threshold: float = 300.0) -> None:
        """
        Create epochs from the continuous data with amplitude rejection criteria.
        
        Parameters
        ----------
        tmin : float
            Start time of epoch in seconds
        tmax : float
            End time of epoch in seconds
        baseline : tuple or None
            Baseline period (start, end) in seconds. None for no baseline correction
        amplitude_threshold : float
            Threshold for rejecting epochs based on peak-to-peak amplitude in µV.
            Default is 300 µV.
        """
        # Convert µV to V for MNE
        reject = dict(eeg=amplitude_threshold * 1e-6)
                
        self.events, self.event_id = mne.events_from_annotations(self.raw)
        
        self.epochs = mne.Epochs(self.raw, 
                            self.events, 
                            event_id=self.event_id,
                            tmin=tmin, 
                            tmax=tmax, 
                            baseline=baseline,
                            reject=reject,
                            reject_by_annotation=True,
                            detrend=0,
                            preload=True,
                            verbose=True)
        
        print(f"Created {len(self.epochs)} epochs")
        if len(self.events) > len(self.epochs):
            n_rejected = len(self.events) - len(self.epochs)
            print(f"Rejected {n_rejected} epochs based on {amplitude_threshold}µV amplitude threshold")
        self.preproc_stats['n_orig_events'] = len(self.events)
        self.preproc_stats['n_final_events'] = len(self.epochs)


    def _get_events(self):
        """Get events from epochs or raw data."""
        if self.epochs is not None:
            return self.epochs.events
        elif hasattr(self, 'raw'):
            return mne.find_events(self.raw, stim_channel='STI 014')
        return None

    def _get_event_ids(self):
        """Get event IDs from epochs or raw data."""
        if self.epochs is not None:
            return self.epochs.event_id
        elif hasattr(self, 'raw'):
            _, event_id = mne.events_from_annotations(self.raw, event_id='auto')
            return event_id
        return None
    
    def clean_muscle_artifacts(self,
                         muscle_window: Tuple[float, float] = (0.005, 0.05),
                         threshold_factor: float = 5.0,
                         n_components: int = 2,
                         verbose: bool = True) -> None:
        """
        Clean TMS-evoked muscle artifacts using tensor decomposition.
        
        Parameters
        ----------
        muscle_window : tuple
            Time window for detecting muscle artifacts in seconds [start, end]
        threshold_factor : float
            Threshold for artifact detection
        n_components : int
            Number of components to use in tensor decomposition
        verbose : bool
            Whether to print progress information
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before cleaning muscle artifacts")
            
        # Create cleaner instance
        cleaner = TMSArtifactCleaner(self.epochs, verbose=verbose)
        
        # Detect artifacts
        artifact_info = cleaner.detect_muscle_artifacts(
            muscle_window=muscle_window,
            threshold_factor=threshold_factor,
            verbose=verbose
        )
        
        if verbose:
            print("\nArtifact detection results:")
            print(f"Found {artifact_info['muscle']['stats']['n_detected']} artifacts")
            print(f"Detection rate: {artifact_info['muscle']['stats']['detection_rate']*100:.1f}%")
        
        # Clean artifacts
        cleaned_epochs = cleaner.clean_muscle_artifacts(
            n_components=n_components,
            verbose=verbose
        )
        
        # Update epochs with cleaned data
        self.epochs = cleaned_epochs
        
        # Apply baseline correction again
        #self.apply_baseline_correction()
        
        if verbose:
            print("\nMuscle artifact cleaning complete")

    def remove_bad_channels(self, threshold: int = 2) -> None:
        """
        Remove and interpolate bad channels using FASTER algorithm.
        
        Parameters
        ----------
        threshold : float
            Threshold for bad channel detection (default = 2)
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before removing bad channels")
            
        bad_channels = find_bad_channels(self.epochs, thres=threshold)
        
        if bad_channels:
            print(f"Detected bad channels: {bad_channels}")
            self.epochs.info['bads'] = list(set(self.epochs.info['bads']).union(set(bad_channels)))
            
            try:
                # First try normal interpolation
                self.epochs.interpolate_bads(reset_bads=True)
                print("Interpolated bad channels")
                
            except ValueError as e:
                print(f"Warning: Standard interpolation failed: {str(e)}")
                print("Attempting alternative interpolation method...")
                
                try:
                    # Try setting montage again with default positions
                    temp_montage = mne.channels.make_standard_montage('standard_1020')
                    self.epochs.set_montage(temp_montage, match_case=False, on_missing='warn')
                    
                    # Try interpolation again
                    self.epochs.interpolate_bads(reset_bads=True)
                    print("Successfully interpolated bad channels using default montage")
                    
                except Exception as e2:
                    print(f"Warning: Alternative interpolation also failed: {str(e2)}")
                    print("Dropping bad channels instead of interpolating")
                    self.epochs.drop_channels(bad_channels)
                    print(f"Dropped channels: {bad_channels}")
            
            self.preproc_stats['bad_channels'] = bad_channels
        else:
            print("No bad channels detected")

    def remove_bad_epochs(self, threshold: int = 3) -> None:
        """
        Remove bad epochs using FASTER algorithm.
        
        Parameters
        ----------
        threshold : float
            Threshold for bad epoch detection
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before removing bad epochs")
            
        bad_epochs = find_bad_epochs(self.epochs, thres=threshold)
        
        if bad_epochs:
            print(f"Dropping {len(bad_epochs)} bad epochs")
            self.epochs.drop(bad_epochs)
            self.preproc_stats['n_bad_epochs'] = len(bad_epochs)
        else:
            print("No bad epochs detected")

    def remove_tms_artifact(self, 
                        cut_times_tms: Tuple[float, float] = (-2, 10), 
                        replace_times: Optional[Tuple[float, float]] = None,
                        verbose: bool = True) -> None:
        """
        Remove TMS artifacts following TESA implementation.
        
        Parameters
        ----------
        cut_times_tms : tuple
            Time window to cut around TMS pulse in ms [start, end]
            Default is [-2, 10] following TESA
        replace_times : tuple, optional
            Time window for calculating average to replace removed data in ms [start, end]
            If None (default), data will be replaced with 0s
        """
        raw_out = self.raw.copy()
        data = raw_out.get_data()
        sfreq = raw_out.info['sfreq']
        
        # Store original info about cut (like TESA's EEG.tmscut)
        if not hasattr(self, 'tmscut'):
            self.tmscut = []
        
        tmscut_info = {
            'cut_times_tms': cut_times_tms,
            'replace_times': replace_times,
            'sfreq': sfreq,
            'interpolated': 'no'
        }
        
        cut_samples = np.round(np.array(cut_times_tms) * sfreq / 1000).astype(int)
        
        # Get TMS annotations
        tms_annotations = [ann for ann in raw_out.annotations 
                        if ann['description'] == 'Stimulation']
        
        print(f"\nFound {len(tms_annotations)} TMS events to process")
        print(f"Removing artifact in window {cut_times_tms} ms")

        processed_count = 0
        skipped_count = 0
        
        for ann in tms_annotations:
            event_sample = int(ann['onset'] * sfreq)
            start = event_sample + cut_samples[0]
            end = event_sample + cut_samples[1]
            
            if start < 0 or end >= data.shape[1]:
                skipped_count += 1
                continue
                
            if replace_times is None:
                data[:, start:end] = 0
            else:
                # Calculate average from replace_times window
                replace_samples = np.round(np.array(replace_times) * sfreq / 1000).astype(int)
                baseline_start = event_sample + replace_samples[0]
                baseline_end = event_sample + replace_samples[1]
                if baseline_start >= 0 and baseline_end < data.shape[1]:
                    baseline_mean = np.mean(data[:, baseline_start:baseline_end], axis=1)
                    data[:, start:end] = baseline_mean[:, np.newaxis]
            processed_count += 1
        
        print(f"Successfully removed artifacts from {processed_count} events")
        if skipped_count > 0:
            print(f"Skipped {skipped_count} events due to window constraints")
        
        raw_out._data = data
        raw_out.set_annotations(raw_out.annotations)
        self.raw = raw_out
        self.tmscut.append(tmscut_info)

    def interpolate_tms_artifact(self, 
                            method: str = 'cubic',
                            interp_window: float = 1.0,
                            cut_times_tms: Tuple[float, float] = (-2, 10),  # Add this back
                            verbose: bool = True) -> None:
        """
        Interpolate TMS artifacts following TESA implementation.
        Uses polynomial interpolation rather than spline interpolation.
        
        Parameters
        ----------
        method : str
            Interpolation method: must be 'cubic' for TESA compatibility
        interp_window : float
            Time window (in ms) before and after artifact for fitting cubic function
            Default is 1.0 ms following TESA
        cut_times_tms : tuple
            Time window where TMS artifact was removed in ms [start, end]
            Default is (-2, 10) following TESA
        verbose : bool
            Whether to print progress information
        """
        if not hasattr(self, 'tmscut') or not self.tmscut:
            raise ValueError("Must run remove_tms_artifact first")
        
        print(f"\nStarting interpolation with {method} method")
        print(f"Using interpolation window of {interp_window} ms")
        print(f"Processing cut window {cut_times_tms} ms")

        interpolated_count = 0
        warning_count = 0
            
        raw_out = self.raw.copy()
        data = raw_out.get_data()
        sfreq = raw_out.info['sfreq']
        
        cut_samples = np.round(np.array(cut_times_tms) * sfreq / 1000).astype(int)
        interp_samples = int(round(interp_window * sfreq / 1000))
            
        for tmscut in self.tmscut:
            if tmscut['interpolated'] == 'no':
                cut_times = tmscut['cut_times_tms']
                cut_samples = np.round(np.array(cut_times) * sfreq / 1000).astype(int)
                interp_samples = int(round(interp_window * sfreq / 1000))
                
                # Process annotations
                tms_annotations = [ann for ann in raw_out.annotations 
                                if ann['description'] == 'Stimulation']
                
                for ann in tms_annotations:
                    event_sample = int(ann['onset'] * sfreq)
                    start = event_sample + cut_samples[0]
                    end = event_sample + cut_samples[1]
                    
                    # Calculate fitting windows
                    window_start = start - interp_samples
                    window_end = end + interp_samples
                    
                    if window_start < 0 or window_end >= data.shape[1]:
                        warning_count += 1
                        continue
                    
                    # Get time points for fitting
                    x = np.arange(window_end - window_start + 1)
                    x_fit = np.concatenate([
                        x[:interp_samples],
                        x[-interp_samples:]
                    ])
                    
                    # Center x values at 0 to avoid badly conditioned warnings (TESA approach)
                    x_fit = x_fit - x_fit[0]
                    if len(x) <= 2 * interp_samples:
                        print(f"Warning: Window too small for interpolation at sample {event_sample}")
                        warning_count += 1
                        continue

                    x_interp = x[interp_samples:-interp_samples] - x_fit[0]

                    # Interpolate each channel using polynomial fit
                    for ch in range(data.shape[0]):
                        y_full = data[ch, window_start:window_end+1]
                        y_fit = np.concatenate([
                            y_full[:interp_samples],
                            y_full[-interp_samples:]
                        ])
                        
                        # Use polynomial fit (like TESA) instead of spline
                        p = np.polyfit(x_fit, y_fit, 3)
                        data[ch, start:end+1] = np.polyval(p, x_interp)
                    
                    interpolated_count += 1


                
                tmscut['interpolated'] = 'yes'
        
        print(f"\nSuccessfully interpolated {interpolated_count} events")
        if warning_count > 0:
            print(f"Encountered {warning_count} warnings during interpolation")
        print("TMS artifact interpolation complete")
        raw_out._data = data
        raw_out.set_annotations(raw_out.annotations)
        self.raw = raw_out

    '''def _apply_interpolation(self, data, epoch_idx, start_idx, end_idx, method, sfreq, window_samples):
        if method == 'linear':
            x_fit = np.array([max(0, start_idx-1), min(data.shape[2 if epoch_idx is not None else 1]-1, end_idx+1)])
            x_interp = np.arange(start_idx, end_idx+1)
            
            for ch in range(data.shape[1]):
                y_fit = data[epoch_idx, ch, x_fit.astype(int)] if epoch_idx is not None else data[ch, x_fit.astype(int)]
                f = interp1d(x_fit, y_fit, kind='linear')
                if epoch_idx is not None:
                    data[epoch_idx, ch, start_idx:end_idx+1] = f(x_interp)
                else:
                    data[ch, start_idx:end_idx+1] = f(x_interp)
        else:  # cubic
            window_start = max(0, start_idx - window_samples[0])
            window_end = min(data.shape[2 if epoch_idx is not None else 1]-1, end_idx + window_samples[1])
            
            x_fit = np.concatenate([
                np.arange(window_start, start_idx),
                np.arange(end_idx+1, window_end+1)
            ])
            x_interp = np.arange(start_idx, end_idx+1)
            
            for ch in range(data.shape[1]):
                y_fit = data[epoch_idx, ch, x_fit.astype(int)] if epoch_idx is not None else data[ch, x_fit.astype(int)]
                x_norm = x_fit - x_fit[0]
                x_interp_norm = x_interp - x_fit[0]
                f = interp1d(x_norm, y_fit, kind='cubic')
                if epoch_idx is not None:
                    data[epoch_idx, ch, start_idx:end_idx+1] = f(x_interp_norm)
                else:
                    data[ch, start_idx:end_idx+1] = f(x_interp_norm)'''
    
    def staged_artifact_removal(self, verbose: bool = True) -> None:
        """
        Perform staged TMS artifact removal following TESA pipeline.
        """
        # Stage 1: Initial removal (-2 to 10 ms)
        if verbose:
            print("Stage 1: Initial TMS artifact removal (-2 to 10 ms)")
        self.remove_tms_artifact(cut_times_tms=(-2, 10), verbose=verbose)
        self.interpolate_tms_artifact(method='cubic', 
                                    interp_window=1.0,
                                    cut_times_tms=(-2, 10),
                                    verbose=verbose)
        
        self.run_ica()
        
        # Stage 2: Extended removal (-2 to 15 ms)
        if verbose:
            print("Stage 2: Extended TMS artifact removal (-2 to 15 ms)")
        self.remove_tms_artifact(cut_times_tms=(-2, 15), verbose=verbose)
        self.interpolate_tms_artifact(method='cubic',
                                    interp_window=5.0,
                                    cut_times_tms=(-2, 15),
                                    verbose=verbose)

        if verbose:
            print("Staged artifact removal complete")

    def run_ica(self, output_dir: str, session_name: str, method: str = "fastica", tms_muscle_thresh: float = 1.0, plot_components: bool = True,) -> None:
        """
        Run first ICA decomposition focusing on TMS-evoked muscle artifacts.
        
        Parameters
        ----------
        method : str
            ICA method to use ('infomax' or 'fastica')
        tms_muscle_thresh : float
            Threshold for muscle component detection
            Default is 1.0 based on empirical testing
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before running ICA")
        # Store pre-ICA epochs
        self.epochs_pre_ica = self.epochs.copy()
        # Set up ICA
        self.ica = ICA(
            max_iter="auto",
            method=method,
            random_state=42
        )
        # Fit ICA
        self.ica.fit(self.epochs)
        # Detect muscle components
        muscle_comps, scores = self.detect_tms_muscle_components(
            tms_muscle_thresh=tms_muscle_thresh,
            verbose=True
        )
        self.preproc_stats['muscle_components'] = muscle_comps
        if plot_components:
            # Plot results
            self.plot_muscle_components(output_dir, session_name, muscle_comps, scores)
        # Apply ICA
        if len(muscle_comps) > 0:
            print(f"Excluding {len(muscle_comps)} muscle components: {muscle_comps}")
            self.ica.apply(self.epochs, exclude=muscle_comps)
        else:
            print("No muscle components detected to exclude")
        # Apply baseline correction maybe not required if we run second ICA
        #self.epochs.apply_baseline(baseline=(-0.1, -0.002))
        #print('Applied baseline corrections post first ICA')

    def filter_raw(self, l_freq=1, h_freq=90, notch_freq=50, notch_width=2, plot_psd=True, fmax=200):
        """
        Apply bandpass and notch filters to raw data.
        
        Parameters
        ----------
        l_freq : float
            Lower cutoff frequency for the bandpass filter
        h_freq : float
            Upper cutoff frequency for the bandpass filter 
        notch_freq : float
            Frequency for the notch filter
        notch_width : float
            Width of the notch filter
        plot_psd : bool
            Whether to plot power spectral density before and after filtering
        fmax: int
            Maximum frequency to display
        """
        if plot_psd:
            # Store copy for before/after comparison
            raw_orig = self.raw.copy()
            
            # Plot PSD before filtering
            fig = raw_orig.compute_psd(fmax=fmax).plot(show=False)
            fig.suptitle('PSD Before Filtering')
        
        # Apply bandpass filter
        self.raw.filter(
            l_freq=l_freq, 
            h_freq=h_freq,
            picks='eeg',
            filter_length='auto',
            l_trans_bandwidth=0.1,
            h_trans_bandwidth=0.5,
            method='fir',
            fir_design='firwin',
            phase='zero',
            verbose=True
        )
        
        # Apply notch filter
        self.raw.notch_filter(
            freqs=notch_freq,
            picks='eeg',
            notch_widths=notch_width,
            filter_length='auto',
            method='fir',
            phase='zero',
            verbose=True
        )
        
        if plot_psd:
            # Plot PSD after filtering
            fig = self.raw.compute_psd(fmax=fmax).plot(show=False)
            fig.suptitle('PSD After Filtering')
            plt.show()
            
        print("Raw data filtering complete")

    def filter_epochs(self, l_freq=1, h_freq=90, notch_freq=50, notch_width=2, notch=True):
        """
        Filter epochs and plot PSD before and after filtering.

        Parameters
        ----------
        l_freq : float
            Lower cutoff frequency for the bandpass filter
        h_freq : float
            Upper cutoff frequency for the bandpass filter 
        notch_freq : float
            Frequency for the notch filter
        notch_width : float
            Width of the notch filter
        notch: bool
            Set to true if you want to apply a notch filter
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before plotting PSD")

        # Before filtering (store copy)
        epochs_orig = self.epochs.copy()

        # Plot PSD before filtering
        fig = epochs_orig.compute_psd(method="multitaper", fmin=0, fmax=100).plot(average=True, show=False)
        fig.suptitle('PSD Before Filtering')

        # Apply bandpass filter
        self.epochs.filter(l_freq=l_freq, 
                        h_freq=h_freq,
                        picks='eeg',
                        filter_length='auto',  
                        l_trans_bandwidth=0.5,  # Specify transition bandwidth for low cutoff
                        h_trans_bandwidth=0.5,  # Specify transition bandwidth for high cutoff
                        method='fir',
                        fir_design='firwin',
                        phase='zero',
                        verbose=True)
        if notch:
            # Apply notch filter
            self.epochs._data = mne.filter.notch_filter(
                self.epochs._data, 
                Fs=self.epochs.info['sfreq'], 
                freqs=notch_freq,
                notch_widths=notch_width, 
                method='fir',
                filter_length='auto',
                phase='zero',
                verbose=True
            )

        # Plot PSD after filtering
        fig = self.epochs.compute_psd(method="multitaper", fmin=0, fmax=100).plot(average=True, show=False)
        fig.suptitle('PSD After Filtering')

        plt.show()

    def run_second_ica(self, method: str = "infomax", 
                    exclude_labels: List[str] = ["eye blink", "muscle artifact", "heart beat", "channel noise"]) -> None:
        """
        Run a second ICA specifically for remaining physiological artifacts using ICLabel.
        
        Parameters
        ----------
        method : str
            ICA method to use ('infomax' or 'fastica')
        exclude_labels : list
            List of ICLabel categories to exclude
        """
        if not hasattr(self, 'epochs'):
            raise ValueError("Must have epochs data to run ICA")
        
        self.set_average_reference()
        if method == "infomax":
            fit_params=dict(extended=True)
        else:
            fit_params=None
        # Set up ICA
        self.ica2 = ICA(
            max_iter="auto",
            method=method,
            random_state=42,
            fit_params=fit_params,
        )
        
        # Fit ICA
        self.ica2.fit(self.epochs)
        
        # Label components with ICLabel
        ic_labels = label_components(self.epochs, self.ica2, method="iclabel")
        labels = ic_labels["labels"]
        
        # Print all component labels
        for n, label in enumerate(labels):
            print(f"Component {n}: {label}")
        
        # Find components to exclude based on ICLabel
        exclude_idx = [
            idx for idx, label in enumerate(labels) if label in exclude_labels
        ]
        
        if len(exclude_idx) > 0:
            print(f"Excluding {len(exclude_idx)} components in second ICA")
            print(f"Excluded components: {exclude_idx}")
            self.ica2.apply(self.epochs, exclude=exclude_idx)
            self.preproc_stats['excluded_ica_components'] = exclude_idx

        else:
            print("No components excluded in second ICA")

        self.epochs.apply_baseline(baseline=(-0.1, -0.002))
        print('Applied baseline corrections post second ICA')

    def detect_tms_muscle_components(self, 
                                tms_muscle_window: Tuple[float, float] = (11, 30), 
                                tms_muscle_thresh: float = 2,  
                                plot_window: Tuple[float, float] = (-200, 500),  
                                verbose: bool = True) -> Tuple[List[int], Dict]:
        """
        Detect TMS-evoked muscle artifacts in ICA components.
        Python implementation of TESA's tesa_compselect muscle detection.
        
        Parameters
        ----------
        tms_muscle_window : tuple
            Time window (in ms) for detecting TMS-evoked muscle activity [start, end]
            Default is (11, 30) ms post-TMS
        tms_muscle_thresh : float
            Threshold for detecting muscle components
            Ratio of mean absolute z-score in muscle window vs. entire time course
        plot_window : tuple
            Time window for plotting in ms [start, end]
        verbose : bool
            Whether to print verbose output
            
        Returns
        -------
        muscle_components : list
            Indices of components classified as muscle artifacts
        component_scores : dict
            Dictionary containing scores and metrics for each component
        """
        if not hasattr(self, 'ica'):
            raise ValueError("Must run ICA before detecting muscle components")
            
        if verbose:
            print(f"Analyzing {self.ica.n_components_} components")
            print(f"Using muscle window: {tms_muscle_window}ms")
            print(f"Using threshold: {tms_muscle_thresh}")
        
        # Get ICA components
        components = self.ica.get_sources(self.epochs)
        n_components = components.get_data().shape[1]
        
        # Get sampling rate and convert windows to samples
        sfreq = self.epochs.info['sfreq']
        tms_muscle_window_samples = np.array([np.abs(self.epochs.times - w/1000).argmin() 
                                            for w in tms_muscle_window])
        
        # Initialize outputs
        muscle_components = []
        component_scores = {
            'muscle_ratios': np.zeros(n_components),
            'window_scores': np.zeros(n_components),
            'total_scores': np.zeros(n_components),
        }
        
        # Process each component
        for comp_idx in range(n_components):
            if verbose and comp_idx % 5 == 0:
                print(f"\nAnalyzing component {comp_idx}")
                
            # Get component time course
            comp_data = components.get_data()[:, comp_idx, :].reshape(-1)
            
            # Z-score the data
            comp_z = zscore(comp_data)
            
            # Calculate scores
            muscle_score = np.abs(comp_z)
            window_score = np.mean(muscle_score[tms_muscle_window_samples[0]:tms_muscle_window_samples[1]])
            total_score = np.mean(muscle_score)
            muscle_ratio = window_score / total_score
            
            if verbose:
                print(f"Comp. {comp_idx} TMS-evoked muscle ratio is {muscle_ratio:.2f}.")
                
            # Store metrics
            component_scores['muscle_ratios'][comp_idx] = muscle_ratio
            component_scores['window_scores'][comp_idx] = window_score
            component_scores['total_scores'][comp_idx] = total_score
            
            # Classify component
            if muscle_ratio >= tms_muscle_thresh:
                muscle_components.append(comp_idx)
        
        return muscle_components, component_scores

    def plot_muscle_components(self, output_dir: str, session_name: str, muscle_components: List[int], 
                            component_scores: Dict,
                            plot_window: Tuple[float, float] = (-100, 250)) -> None:
        """
        Plot detected muscle components using MNE's plotting functions.
        
        Parameters
        ----------
        muscle_components : list
            Indices of components classified as muscle artifacts
        component_scores : dict
            Dictionary containing component metrics 
        plot_window : tuple
            Time window for plotting in ms [start, end]
        """
        if not hasattr(self, 'ica'):
            raise ValueError("Must run ICA before plotting components")
            
        if len(muscle_components) == 0:
            print("No muscle components detected to plot")
            
            # Plot muscle ratios
            plt.figure(figsize=(10, 4))
            plt.bar(range(len(component_scores['muscle_ratios'])), 
                    component_scores['muscle_ratios'])
            plt.axhline(y=2, color='r', linestyle='--', label='Threshold')
            plt.xlabel('Component')
            plt.ylabel('Muscle Ratio')
            plt.title('TMS-Evoked Muscle Activity')
            plt.legend()
            plt.show()
            return
        with mne.viz.use_browser_backend("matplotlib"):
            fig = self.ica.plot_sources(self.epochs, show=False)
        # Plot detected muscle components

        fig.savefig(os.path.join(output_dir, session_name, f"Muscle components.png"), 
            dpi=300, bbox_inches='tight')
        plt.close(fig)
        
        '''
        # Plot component properties 
        for comp in muscle_components:
            self.ica.plot_properties(self.epochs, picks=comp, 
                                psd_args=dict(fmax=70))
        
        # Create summary plot
        fig, ax = plt.subplots(figsize=(12, 4))
        
        # Plot muscle ratios
        bars = ax.bar(range(len(component_scores['muscle_ratios'])), 
                    component_scores['muscle_ratios'])
        ax.axhline(y=2, color='r', linestyle='--', label='Threshold')
        
        # Color the muscle components
        for comp in muscle_components:
            bars[comp].set_color('red')
        
        ax.set_xlabel('Component')
        ax.set_ylabel('Muscle Ratio')
        ax.set_title('TMS-Evoked Muscle Activity')
        ax.legend()
        
        plt.tight_layout()
        plt.show()
        '''
    def set_average_reference(self):
        '''
        - Rereference EEG and apply projections
        '''
        self.epochs.set_eeg_reference('average', projection=True)
        print("Rereferenced epochs to 'average'")

    def apply_baseline_correction(self, baseline: Tuple[float, float] = (-0.1, -0.002)) -> None:
        """
        Apply baseline correction to epochs.
        
        Parameters
        ----------
        baseline : tuple
            Start and end time of baseline period in seconds (start, end)
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before applying baseline")
            
        self.epochs.apply_baseline(baseline=baseline)
        print(f"Applied baseline correction using window {baseline} seconds")

    def downsample(self):
        '''
        - Downsample epochs to desired sfreq if current sfreq > desired sfreq (default 1000 Hz)
        '''

        current_sfreq = self.epochs.info['sfreq']
        if current_sfreq > self.ds_sfreq:
            self.epochs = self.epochs.resample(self.ds_sfreq)
            print(f"Downsampled data to {self.ds_sfreq} Hz")
        else:
            print("Current sfreq < target sfreq")
            pass

    def get_preproc_stats(self):
        """Return current preprocessing statistics"""
        return {
            'Original Events': self.preproc_stats['n_orig_events'],
            'Final Events': self.preproc_stats['n_final_events'],
            'Event Retention Rate': f"{(self.preproc_stats['n_final_events']/self.preproc_stats['n_orig_events'])*100:.1f}%",
            'Bad Channels': ', '.join(self.preproc_stats['bad_channels']) if self.preproc_stats['bad_channels'] else 'None',
            'Bad Epochs Removed': self.preproc_stats['n_bad_epochs'],
            'ICA1 Muscle Components': len(self.preproc_stats['muscle_components']),
            'ICA2 Excluded Components': len(self.preproc_stats['excluded_ica_components']),
            'TMS Interpolation Windows': len(self.preproc_stats['interpolated_times'])
    }


    def plot_evoked_response(self, picks: Optional[str] = None,
                            ylim: Dict = {'eeg': [-7, 7]},
                            xlim: Optional[Tuple[float, float]] = (-0.1, 0.3),
                            title: str = 'Evoked Response',
                            show: bool = True) -> None:
        """
        Plot averaged evoked response with butterfly plot and global field power.
        
        Parameters
        ----------
        picks : str or list
            Channels to include in plot
        ylim : dict
            Y-axis limits for different channel types
        xlim : tuple
            X-axis limits in seconds (start_time, end_time)
        title : str
            Title for the plot
        show : bool
            Whether to show the plot immediately
        """
        if self.epochs is None:
            raise ValueError("Must create epochs before plotting evoked response")
            
        # Create evoked from epochs
        evoked = self.epochs.average()
        
        # Create figure with two subplots
        fig = plt.figure(figsize=(12, 8))
        gs = plt.GridSpec(2, 1, height_ratios=[3, 1])
        
        # Butterfly plot
        ax1 = fig.add_subplot(gs[0])
        evoked.plot(picks=picks, axes=ax1, ylim=ylim, xlim=xlim, show=False)
        ax1.set_title(f'{title} - Butterfly Plot')
        
        # GFP plot
        ax2 = fig.add_subplot(gs[1])
        gfp = np.std(evoked.data, axis=0) * 1e6  # Convert to μV
        times = evoked.times
        ax2.plot(times, gfp, 'b-', linewidth=2)
        ax2.set_xlabel('Time (s)')
        ax2.set_ylabel('GFP (μV)')
        ax2.set_title('Global Field Power')
        ax2.grid(True)
        if xlim is not None:
            ax2.set_xlim(xlim)
        
        # Add vertical line at t=0
        for ax in [ax1, ax2]:
            ax.axvline(x=0, color='r', linestyle='--', alpha=0.5)
            
        plt.tight_layout()
        
        if show:
            plt.show()
        
        return fig


    def apply_ssp(self, n_eeg=2):

        
        projs_epochs = mne.compute_proj_epochs(self.epochs, n_eeg=n_eeg, n_jobs=-1, verbose=True)
        self.epochs.add_proj(projs_epochs)
        self.epochs.apply_proj()
        
        
    def plot_epochs(self, ylim: Optional[Dict] = None) -> None:
        """
        Plot epochs with inverted y-axis.
        
        Parameters
        ----------
        ylim : dict or None
            Y-axis limits for plotting
        """
        if self.epochs is None:
            raise ValueError("No epochs available to plot")
            
        with mne.viz.use_browser_backend("matplotlib"):
            fig = self.epochs.copy().plot()
            for ax in fig.get_axes():
                if hasattr(ax, 'invert_yaxis'):
                    ax.invert_yaxis()
            fig.canvas.draw()
            
    def plot_ica_components(self) -> None:
        """
        Plot ICA components if ICA has been run.
        """
        if not hasattr(self, 'ica'):
            raise ValueError("Must run ICA before plotting components")
            
        with mne.viz.use_browser_backend("matplotlib"):    
            self.ica.plot_sources(self.epochs, show_scrollbars=True)



    def apply_csd(self, lambda2=1e-5, stiffness=4, n_legendre_terms=50, verbose=True):
        """
        Apply Current Source Density transformation maintaining CSD channel type.
        
        Parameters
        ----------
        lambda2 : float
            Regularization parameter
        stiffness : int
            Stiffness of the spline
        n_legendre_terms : int
            Number of Legendre terms
        verbose : bool
            Print progress information
        """
        if verbose:
            print("Applying Current Source Density transformation...")
        
        # Apply CSD transformation
        self.epochs = compute_current_source_density(
            self.epochs,
            lambda2=lambda2,
            stiffness=stiffness,
            n_legendre_terms=n_legendre_terms,
            copy=True
        )
        
        # The channels are now CSD type, so we leave them as is
        if verbose:
            print("CSD transformation complete")
        
        # Store the fact that we've applied CSD
        self.csd_applied = True
        
        return self.epochs
            

    def fix_tms_artifact(self, 
                           window: Tuple[float, float] = (-0.002, 0.015),
                           mode: str = 'window') -> None:
        """
        Interpolate the TMS artifact using MNE's fix_stim_artifact function.
        
        Parameters
        ----------
        window : tuple
            Time window around TMS pulse to interpolate (start, end) in seconds
        mode : str
            Interpolation mode ('linear', 'cubic', or 'hann')
        """
        if self.raw is None:
            raise ValueError("Must create raw before interpolating TMS artifact")
        
        events, event_id = mne.events_from_annotations(self.raw)
        
        try:
            self.raw = mne.preprocessing.fix_stim_artifact(
                self.raw,
                events=events,
                event_id=event_id,
                tmin=window[0],
                tmax=window[1],
                mode=mode
            )
            print(f"Applied TMS artifact interpolation with mode '{mode}'")
        except Exception as e:
            print(f"Error in TMS artifact interpolation: {str(e)}")


    def downsample(self, target_sfreq: float = None):
        """
        Downsample epochs to the desired sampling frequency if the current sampling frequency is higher.

        Parameters
        ----------
        target_sfreq : float or None
            The target sampling frequency. If None, uses self.ds_sfreq.
        """
        if target_sfreq is None:
            target_sfreq = self.ds_sfreq
        current_sfreq = self.epochs.info['sfreq']
        if current_sfreq > target_sfreq:
            self.epochs = self.epochs.resample(target_sfreq)
            print(f"Downsampled data to {target_sfreq} Hz")
        else:
            print("Current sfreq <= target sfreq; no downsampling performed")
    
    def save_epochs(self, fpath: str = None):
        """
        Save preprocessed epochs
        """
        self.epochs.save(fpath, verbose=True, overwrite=True)

        print(f"Epochs saved at {fpath}")



from numpy.typing import ArrayLike, NDArray

from .utils import _format_input as _format_input

def _format_input_for_onnx(topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike):
    """Format the features to the correct shape and type for ONNX."""

def _run_iclabel(images: ArrayLike, psds: ArrayLike, autocorr: ArrayLike) -> NDArray:
    """Run ICLabel using onnx."""



# gui_app.py
import tkinter as tk
from tkinter import ttk, filedialog, messagebox
import threading
from pathlib import Path
import sys
import io
import os

# When importing from run.py, update to:
from ..run import process_subjects

class ToolTip(object):
    def __init__(self, widget, text='widget info'):
        self.widget = widget
        self.text = text
        self.tooltip = None
        self.id = None  # For scheduled showing
        self.widget.bind("<Enter>", self.schedule)
        self.widget.bind("<Leave>", self.leave)

    def schedule(self, event=None):
        self.unschedule()
        self.id = self.widget.after(500, self.enter)  # 500ms delay

    def unschedule(self):
        if self.id:
            self.widget.after_cancel(self.id)
            self.id = None

    def enter(self, event=None):
        x, y, _, _ = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        self.tooltip = tk.Toplevel(self.widget)
        self.tooltip.wm_overrideredirect(True)
        self.tooltip.wm_geometry(f"+{x}+{y}")
        label = ttk.Label(self.tooltip, text=self.text, justify='left',
                         background="#ffffe0", relief='solid', borderwidth=1)
        label.pack()

    def leave(self, event=None):
        self.unschedule()
        if self.tooltip:
            self.tooltip.destroy()
            self.tooltip = None

class TMSEEG_GUI:
    def __init__(self, root):
        self.root = root
        self.root.title("TMS-EEG Preprocessing and PCIst")
        
        # Set initial window size (width x height)
        self.root.geometry("1000x1000")
        
        # Initialize state variables first
        self.show_advanced = tk.BooleanVar(value=False)
        self.running = False
        
        # Initialize validation rules and tooltips
        self.init_validation_rules()
        self.init_tooltips()
        
        # Create scroll canvas
        self.create_scroll_canvas()
        
        # Create GUI elements
        self.create_gui_elements()

    def init_validation_rules(self):
        self.validation_rules = {
            'ds_sfreq': {'min': 100, 'max': 5000, 'type': float},
            'random_seed': {'min': 0, 'max': 1000000, 'type': int},
            'bad_channels_threshold': {'min': 0, 'max': 10, 'type': float},
            'bad_epochs_threshold': {'min': 0, 'max': 10, 'type': float},
            'amplitude_threshold': {'min': 50, 'max': 1000, 'type': float},
            'tms_muscle_thresh': {'min': 0, 'max': 10, 'type': float},
            'n_components': {'min': 1, 'max': 100, 'type': int},
            'ssp_n_eeg': {'min': 1, 'max': 10, 'type': int},
            'l_freq': {'min': 0, 'max': 100, 'type': float},
            'h_freq': {'min': 1, 'max': 1000, 'type': float},
            'notch_freq': {'min': 0, 'max': 1000, 'type': float},
            'notch_width': {'min': 0, 'max': 10, 'type': float},
            'initial_cut_start': {'min': -100, 'max': 0, 'type': float},
            'initial_cut_end': {'min': 0, 'max': 100, 'type': float},
            'initial_interp_window': {'min': 0.1, 'max': 10.0, 'type': float},
            'extended_cut_start': {'min': -100, 'max': 0, 'type': float},
            'extended_cut_end': {'min': 0, 'max': 100, 'type': float},
            'extended_interp_window': {'min': 0.1, 'max': 10.0, 'type': float},
            'interpolation_method': {'type': 'str', 'choices': ['cubic']},
            'epochs_tmin': {'min': -2, 'max': 0, 'type': float},
            'epochs_tmax': {'min': 0, 'max': 2, 'type': float},
            'baseline_start': {'min': -1000, 'max': 0, 'type': int},
            'baseline_end': {'min': -1000, 'max': 0, 'type': int},
            'k': {'min': 0, 'max': 10, 'type': float},
            'min_snr': {'min': 0, 'max': 10, 'type': float},
            'max_var': {'min': 0, 'max': 100, 'type': float},
            'n_steps': {'min': 10, 'max': 1000, 'type': int},
            'n_clusters': {'min': 2, 'max': 20, 'type': int},
            'n_resamples': {'min': 1, 'max': 100, 'type': int},
            'n_samples': {'min': 100, 'max': 10000, 'type': int},
            'min_peak_distance': {'min': 1, 'max': 100, 'type': int},
            'substitute_zero_events_with': {'min': 1, 'max': 100, 'type': int},
            'threshold_factor': {'min': 0.1, 'max': 10.0, 'type': float},
            'lambda2': {'min': 1e-6, 'max': 1e-4, 'type': float},
            'stiffness': {'min': 1, 'max': 10, 'type': int},
            'ica_method': {'type': 'str', 'choices': ['fastica', 'infomax']},
            'second_ica_method': {'type': 'str', 'choices': ['fastica', 'infomax']},
            'interpolation_method': {'type': 'str', 'choices': ['cubic', 'linear']},
            'response_start': {'min': -1000, 'max': 1000, 'type': int},
            'response_end': {'min': -1000, 'max': 1000, 'type': int},
            'prominence': {'min': 0.01, 'max': 1.0, 'type': float}
            
        }

    def init_tooltips(self):
        self.tooltips = {
            'ds_sfreq': 'Downsampling frequency in Hz',
            'random_seed': 'Random seed for reproducibility',
            'bad_channels_threshold': 'Threshold for detecting bad channels',
            'bad_epochs_threshold': 'Threshold for detecting bad epochs',
            'amplitude_threshold': 'Threshold for epoch rejection based on peak-to-peak amplitude in µV',
            'tms_muscle_thresh': 'Threshold for detecting muscle artifacts',
            'n_components': 'Number of components for PARAFAC',
            'ssp_n_eeg': 'Number of EEG components for SSP',
            'l_freq': 'Lower frequency cutoff for filtering',
            'h_freq': 'Upper frequency cutoff for filtering',
            'notch_freq': 'Frequency for notch filter',
            'notch_width': 'Width of the notch filter',
            'initial_cut_start': 'Start time for initial TMS artifact removal (TESA default: -2)',
            'initial_cut_end': 'End time for initial TMS artifact removal (TESA default: 10)',
            'initial_interp_window': 'Initial interpolation window (TESA default: 1.0)',
            'extended_cut_start': 'Start time for extended TMS artifact removal (TESA default: -2)',
            'extended_cut_end': 'End time for extended TMS artifact removal (TESA default: 15)',
            'extended_interp_window': 'Extended interpolation window (TESA default: 5.0)',
            'interpolation_method': 'Interpolation method (TESA requires cubic)',
            'epochs_tmin': 'Start time for epochs (s)',
            'epochs_tmax': 'End time for epochs (s)',
            'baseline_start': 'Start time for baseline period (ms)',
            'baseline_end': 'End time for baseline period (ms)',
            'k': 'PCIst parameter k',
            'min_snr': 'Minimum signal-to-noise ratio',
            'max_var': 'Maximum variance explained (%)',
            'n_steps': 'Number of steps for PCIst calculation',
            'min_peak_distance': 'Minimum distance between peaks',
            'substitute_zero_events_with': 'Value to substitute zero events with in the data',
            'threshold_factor': 'Threshold factor for muscle artifact cleaning',
            'muscle_window_start': 'Start time for muscle artifact window (s)',
            'muscle_window_end': 'End time for muscle artifact window (s)',
            'ica_method': 'Method for first ICA (fastica or infomax)',
            'second_ica_method': 'Method for second ICA (fastica or infomax)',
            'lambda2': 'Regularization parameter for CSD transformation',
            'stiffness': 'Stiffness parameter for CSD transformation',
            'response_start': 'Start of response window in ms',
            'response_end': 'End of response window in ms',
            'prominence': 'Minimum prominence for peak detection as fraction of max GFP (default: 0.01)'
        }
        
    def create_scroll_canvas(self):
        self.main_canvas = tk.Canvas(self.root)
        scrollbar = ttk.Scrollbar(self.root, orient="vertical", command=self.main_canvas.yview)
        self.scrollable_frame = ttk.Frame(self.main_canvas)
        
        self.scrollable_frame.bind(
            "<Configure>",
            lambda e: self.main_canvas.configure(scrollregion=self.main_canvas.bbox("all"))
        )
        
        self.main_canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")
        self.main_canvas.configure(yscrollcommand=scrollbar.set)
        
        self.main_canvas.pack(side="left", fill="both", expand=True)
        scrollbar.pack(side="right", fill="y")
        
        # Mouse wheel scrolling
        self.root.bind("<MouseWheel>", self._on_mousewheel)
        
    def create_gui_elements(self):
        main_frame = ttk.Frame(self.scrollable_frame, padding="10")
        main_frame.grid(row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))
        
        # Directory selectors (always visible)
        self.create_directory_selectors(main_frame)
        
        # Toggle button for advanced options
        ttk.Checkbutton(main_frame, text="Show Advanced Options", 
                        variable=self.show_advanced,
                        command=self.toggle_advanced_options).grid(
                            row=1, column=0, sticky=tk.W, pady=(10,0))
        
        # Create frames for options (hidden by default)
        self.options_frame = ttk.Frame(main_frame)
        self.options_frame.grid(row=2, column=0, sticky=(tk.W, tk.E), pady=5)
        self.options_frame.grid_remove()  # Hide initially
        
        # Create basic and advanced options in the hidden frame
        self.create_basic_options(self.options_frame)
        self.create_advanced_options(self.options_frame)
        
        # Console always visible at bottom
        self.create_console(main_frame)
        self.create_control_buttons(main_frame)

    def toggle_advanced_options(self):
        if self.show_advanced.get():
            self.options_frame.grid()
        else:
            self.options_frame.grid_remove()

    def create_directory_selectors(self, parent):
        dir_frame = ttk.LabelFrame(parent, text="Data Selection", padding="10")
        dir_frame.grid(row=0, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=10)
        
        # Make directory selectors more prominent
        ttk.Label(dir_frame, text="Data Directory:", font=('TkDefaultFont', 10, 'bold')).grid(
            row=0, column=0, sticky=tk.W, padx=5, pady=5)
        self.data_dir = tk.StringVar(value=str(Path.cwd() / 'data'))
        ttk.Entry(dir_frame, textvariable=self.data_dir, width=60).grid(
            row=0, column=1, padx=5, pady=5)
        ttk.Button(dir_frame, text="Browse", command=self.browse_data_dir).grid(
            row=0, column=2, padx=5, pady=5)
        
        ttk.Label(dir_frame, text="Output Directory:", font=('TkDefaultFont', 10, 'bold')).grid(
            row=1, column=0, sticky=tk.W, padx=5, pady=5)
        self.output_dir = tk.StringVar(value=str(Path.cwd() / 'output'))
        ttk.Entry(dir_frame, textvariable=self.output_dir, width=60).grid(
            row=1, column=1, padx=5, pady=5)
        ttk.Button(dir_frame, text="Browse", command=self.browse_output_dir).grid(
            row=1, column=2, padx=5, pady=5)
        
    def validate_parameter(self, param, value):
        rules = self.validation_rules.get(param)
        if not rules:
            return True
            
        if rules['type'] == 'str':
            return value in rules['choices']
            
        try:
            # First convert to float to handle all numeric inputs
            val = float(value)
            
            # Then convert to int if needed, using round()
            if rules['type'] == int:
                val = round(val)
                
            # Check bounds
            if 'min' in rules and val < rules['min']:
                return False
            if 'max' in rules and val > rules['max']:
                return False
                
            return True
        except ValueError:
            return False

    def add_parameter_with_validation(self, parent, label, param, default, row):
        ttk.Label(parent, text=label + ":").grid(row=row, column=0, sticky=tk.W, padx=5, pady=2)
        var = tk.StringVar(value=str(default))
        entry = ttk.Entry(parent, textvariable=var, width=10)
        entry.grid(row=row, column=1, padx=5, pady=2)
        
        # Add tooltip
        if param in self.tooltips:
            ToolTip(entry, self.tooltips[param])
        
        # Add validation
        def validate(*args):
            if not self.validate_parameter(param, var.get()):
                rules = self.validation_rules[param]
                messagebox.showwarning("Invalid Value", 
                    f"Invalid value for {param}.\n"
                    f"Expected {rules['type'].__name__} between {rules['min']} and {rules['max']}")
                var.set(str(default))
        
        var.trace_add('write', validate)  # Updated from trace() to trace_add()
        return var

    def run_analysis(self):
        if self.running:
            return
            
        # Validate all parameters before running
        for param, var in self.params.items():
            if not self.validate_parameter(param, var.get()):
                rules = self.validation_rules[param]
                messagebox.showerror("Validation Error", 
                    f"Invalid value for {param}.\n"
                    f"Expected {rules['type'].__name__} between {rules['min']} and {rules['max']}")
                return
                
        # Convert parameters to correct types for args
        args_dict = {
            'data_dir': self.data_dir.get(),
            'output_dir': self.output_dir.get(),
            'stim_channel': 'STI 014', 
            'plot_preproc': self.plot_preproc.get(),
            'stim_channel': self.stim_channel.get(), 
            'clean_muscle_artifacts': self.clean_muscle.get(),
            'show_evoked': self.show_evoked.get(),
            'research': self.research_stats.get(),
            #'preproc_qc': self.preproc_qc.get(),
            'apply_ssp': self.apply_ssp.get(),
            'apply_csd': self.apply_csd.get(),
        }
        
        # Add numerical parameters with proper type conversion
        for param, var in self.params.items():
            try:
                value = float(var.get())
                if self.validation_rules[param]['type'] == int:
                    value = round(value)
                args_dict[param] = value
            except (ValueError, KeyError):
                messagebox.showerror("Error", f"Invalid value for {param}")
                self.progress.stop()
                self.running = False
                return
        
        # Create Args object
        class Args:
            pass
        args = Args()
        for key, value in args_dict.items():
            setattr(args, key, value)
            
        self.console.delete(1.0, tk.END)
        self.progress.start()
        self.running = True
        
        # Run analysis in separate thread
        thread = threading.Thread(target=self.run_analysis_thread, args=(args,))
        thread.daemon = True
        thread.start()
        
    def create_parameter_group(self, parent, params, start_row=0):
        vars = {}
        for i, (label, (param, default)) in enumerate(params.items()):
            vars[param] = self.add_parameter_with_validation(parent, label, param, default, start_row + i)
        return vars
    
    def add_parameter_group(self, parent, params):
        for i, (label, (param, default)) in enumerate(params.items()):
            ttk.Label(parent, text=label + ":").grid(row=i, column=0, sticky=tk.W, padx=5, pady=2)
            var = self.add_parameter_with_validation(parent, label, param, default, i)
            self.params[param] = var
    
    def _on_mousewheel(self, event):
        self.main_canvas.yview_scroll(int(-1*(event.delta/120)), "units")
        
    def create_basic_options(self, parent):
        options_frame = ttk.LabelFrame(parent, text="Basic Options", padding="5")
        options_frame.grid(row=1, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=5)
        
        # First row
        self.plot_preproc = tk.BooleanVar()
        plot_button = ttk.Checkbutton(options_frame, text="Plot Preprocessing Steps", 
                    variable=self.plot_preproc)
        plot_button.grid(row=3, column=1, sticky=tk.W)
        ToolTip(plot_button, "Show preprocessing visualization steps")
        
        self.clean_muscle = tk.BooleanVar()
        ttk.Checkbutton(options_frame, text="Clean Muscle Artifacts with PARAFAC decomposition", 
                    variable=self.clean_muscle).grid(row=0, column=0, sticky=tk.W)
        
        # Second row
        self.show_evoked = tk.BooleanVar()
        ttk.Checkbutton(options_frame, text="Show Evoked Plot", 
                    variable=self.show_evoked).grid(row=3, column=0, sticky=tk.W)
        
        self.research_stats = tk.BooleanVar()
        ttk.Checkbutton(options_frame, text="Generate Research Statistics", 
                    variable=self.research_stats).grid(row=0, column=1, sticky=tk.W)
        
        # Third row
        self.apply_ssp = tk.BooleanVar(value=False)  
        ssp_button = ttk.Checkbutton(options_frame, text="Apply SSP", 
                    variable=self.apply_ssp)
        ssp_button.grid(row=1, column=0, sticky=tk.W)
        ToolTip(ssp_button, "Apply Signal Space Projection for artifact removal")
        
        #self.preproc_qc = tk.BooleanVar()
       # ttk.Checkbutton(options_frame, text="Generate Preprocessing QC", 
                   # variable=self.preproc_qc).grid(row=0, column=1, sticky=tk.W)
        
        # Fourth row
        self.apply_csd = tk.BooleanVar(value=True)  
        csd_button = ttk.Checkbutton(options_frame, text="Apply CSD", 
                    variable=self.apply_csd)
        csd_button.grid(row=2, column=0, sticky=tk.W)
        ToolTip(csd_button, "Apply Current Source Density transformation")

  
        self.validate_teps = tk.BooleanVar()  
        validate_button = ttk.Checkbutton(options_frame, text="Validate TEPs", 
                    variable=self.validate_teps)
        validate_button.grid(row=1, column=1, sticky=tk.W)
        ToolTip(validate_button, "Perform TEP validation and generate reports")
        
        self.save_validation = tk.BooleanVar() 
        save_button = ttk.Checkbutton(options_frame, text="Save Validation Reports & Plots", 
                    variable=self.save_validation)
        save_button.grid(row=2, column=1, sticky=tk.W)
        ToolTip(save_button, "Save validation reports and plots to output directory")

        format_frame = ttk.LabelFrame(options_frame, text="Data Format", padding="5")
        format_frame.grid(row=4, column=0, columnspan=2, sticky=(tk.W, tk.E), pady=5)

        ttk.Label(format_frame, text="Data Format:").grid(row=0, column=0, sticky=tk.W, padx=5)
        self.data_format = tk.StringVar(value='neurone')
        format_combo = ttk.Combobox(format_frame, textvariable=self.data_format, width=15)
        format_combo['values'] = ('neurone', 'brainvision', 'edf', 'cnt', 'eeglab', 'auto')
        format_combo.grid(row=0, column=1, padx=5, pady=2)
        format_combo.state(['readonly'])
        ToolTip(format_combo, "Format of input data files")

        ttk.Label(format_frame, text="EEGLAB Units:").grid(row=1, column=0, sticky=tk.W, padx=5)
        self.eeglab_units = tk.StringVar(value='auto')
        units_combo = ttk.Combobox(format_frame, textvariable=self.eeglab_units, width=15)
        units_combo['values'] = ('auto', 'mm', 'm', 'cm')
        units_combo.grid(row=1, column=1, padx=5, pady=2)
        units_combo.state(['readonly'])
        ToolTip(units_combo, "Units for EEGLAB channel positions")

        ttk.Label(options_frame, text="Stim Channel:").grid(row=5, column=0, sticky=tk.W, padx=5)
        self.stim_channel = tk.StringVar(value='STI 014')
        stim_entry = ttk.Entry(options_frame, textvariable=self.stim_channel, width=15)
        stim_entry.grid(row=5, column=1, padx=5, pady=2)
        ToolTip(stim_entry, "Name of the stimulus channel in your data")
            
        
    def create_advanced_options(self, parent):
        # Notebook for advanced parameters
        notebook = ttk.Notebook(parent)
        notebook.grid(row=2, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=5)
        
        # Create different tabs for parameter categories
        self.params = {}
        
        # Preprocessing Parameters
        preproc_frame = ttk.Frame(notebook, padding="5")
        notebook.add(preproc_frame, text="Preprocessing")
        self.add_parameter_group(preproc_frame, {
            'Downsampling Frequency (Hz)': ('ds_sfreq', 725),
            'Random Seed': ('random_seed', 42),
            'Bad Channels Threshold': ('bad_channels_threshold', 1),
            'Bad Epochs Threshold': ('bad_epochs_threshold', 1),
            'SSP EEG Components': ('ssp_n_eeg', 2),
            'Substitute Zero Events With': ('substitute_zero_events_with', 10),
        })
        
        # Filtering Parameters
        filter_frame = ttk.Frame(notebook, padding="5")
        notebook.add(filter_frame, text="Filtering")
        self.add_parameter_group(filter_frame, {
            'Low Frequency (Hz)': ('l_freq', 1), # 0.1 is the standard in the PCIst article by Comolatti et al., 2019
            'High Frequency (Hz)': ('h_freq', 45),
            'Notch Frequency (Hz)': ('notch_freq', 50),
            'Notch Width': ('notch_width', 2),
        })
        
        # TMS Parameters
        tms_frame = ttk.Frame(notebook, padding="5")
        notebook.add(tms_frame, text="TMS Settings")

        # Create sub-frames for each stage
        stage1_frame = ttk.LabelFrame(tms_frame, text="Stage 1 (Initial Removal)", padding="5")
        stage1_frame.grid(row=0, column=0, sticky=(tk.W, tk.E), pady=5)
        self.add_parameter_group(stage1_frame, {
            'Initial Cut Start (ms)': ('initial_cut_start', -2),
            'Initial Cut End (ms)': ('initial_cut_end', 10),
            'Initial Interp Window (ms)': ('initial_interp_window', 1.0),
        })

        stage2_frame = ttk.LabelFrame(tms_frame, text="Stage 2 (Extended Removal)", padding="5")
        stage2_frame.grid(row=1, column=0, sticky=(tk.W, tk.E), pady=5)
        self.add_parameter_group(stage2_frame, {
            'Extended Cut Start (ms)': ('extended_cut_start', -2),
            'Extended Cut End (ms)': ('extended_cut_end', 15),
            'Extended Interp Window (ms)': ('extended_interp_window', 5.0),
        })

        # Common parameters
        common_frame = ttk.LabelFrame(tms_frame, text="Common Settings", padding="5")
        common_frame.grid(row=2, column=0, sticky=(tk.W, tk.E), pady=5)
        self.add_parameter_group(common_frame, {
            'Interpolation Method': ('interpolation_method', 'cubic')
        })

        self.skip_second_artifact_removal = tk.BooleanVar()
        ttk.Checkbutton(stage2_frame, text="Skip Second TMS Artifact Removal", 
                        variable=self.skip_second_artifact_removal).grid(row=0, column=2, padx=5, pady=2, sticky=tk.W)
        
        # Muscle Artifact Parameters
        muscle_frame = ttk.Frame(notebook, padding="5")
        notebook.add(muscle_frame, text="Muscle Artifacts (PARAFAC)")
        self.add_parameter_group(muscle_frame, {
            'Muscle Window Start (s)': ('muscle_window_start', 0.005),
            'Muscle Window End (s)': ('muscle_window_end', 0.030),
            'Threshold Factor': ('threshold_factor', 1.0),
            'Number of PARAFAC Components': ('n_components', 5),
        })
        
        # ICA Parameters
        ica_frame = ttk.Frame(notebook, padding="5")
        notebook.add(ica_frame, text="ICA Settings")
        self.add_parameter_group(ica_frame, {
            'ICA Method': ('ica_method', 'fastica'),
            'TMS Muscle Threshold (for first ICA)': ('tms_muscle_thresh', 2.0),
            'Second ICA Method': ('second_ica_method', 'infomax'),
        })
        
        # CSD Parameters
        csd_frame = ttk.Frame(notebook, padding="5")
        notebook.add(csd_frame, text="CSD Settings")
        self.add_parameter_group(csd_frame, {
            'Lambda2': ('lambda2', 1e-3),
            'Stiffness': ('stiffness', 3),
        })
        
        # Epoch Parameters
        epoch_frame = ttk.Frame(notebook, padding="5")
        notebook.add(epoch_frame, text="Epoching")
        self.add_parameter_group(epoch_frame, {
            'Epoch Start Time (s)': ('epochs_tmin', -0.41),
            'Epoch End Time (s)': ('epochs_tmax', 0.41),
            'Baseline Start (ms)': ('baseline_start', -400),
            'Baseline End (ms)': ('baseline_end', -50),
            'Amplitude Threshold (µV)': ('amplitude_threshold', 4500), 
        })
        tep_frame = ttk.Frame(notebook, padding="5")
        notebook.add(tep_frame, text="TEP Analysis")
        self.add_parameter_group(tep_frame, {
            'Prominence': ('prominence', 0.01)  
        })
        
        # PCIst Parameters
        pcist_frame = ttk.Frame(notebook, padding="5")
        notebook.add(pcist_frame, text="PCIst")
        self.add_parameter_group(pcist_frame, {
            'k Parameter': ('k', 1.2),
            'Min SNR': ('min_snr', 1.1),
            'Max Variance (%)': ('max_var', 99.0),
            'Number of Steps': ('n_steps', 100),
        })
        
        
    def add_parameter_group(self, parent, params):
        for i, (label, (param, default)) in enumerate(params.items()):
            ttk.Label(parent, text=label + ":").grid(row=i, column=0, sticky=tk.W, padx=5, pady=2)
            self.params[param] = tk.StringVar(value=str(default))
            ttk.Entry(parent, textvariable=self.params[param], width=10).grid(row=i, column=1, padx=5, pady=2)
            
    def create_console(self, parent):
        # Console frame
        console_frame = ttk.LabelFrame(parent, text="Console Output", padding="5")
        console_frame.grid(row=4, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=5)
        
        self.console = tk.Text(console_frame, height=15, width=120)
        self.console.grid(row=0, column=0, sticky=(tk.W, tk.E))
        
        scrollbar = ttk.Scrollbar(console_frame, orient="vertical", command=self.console.yview)
        scrollbar.grid(row=0, column=1, sticky=(tk.N, tk.S))
        
        self.console.configure(yscrollcommand=scrollbar.set)
        
    def create_control_buttons(self, parent):
        button_frame = ttk.Frame(parent)
        button_frame.grid(row=2, column=0, columnspan=3, pady=10)
        
        ttk.Button(button_frame, text="Get PCIst", command=self.run_analysis).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Stop", command=self.stop_analysis).pack(side=tk.LEFT, padx=5)
        
        self.progress = ttk.Progressbar(parent, length=300, mode='indeterminate')
        self.progress.grid(row=3, column=0, columnspan=3, pady=5)
        
    def browse_data_dir(self):
        directory = filedialog.askdirectory(initialdir=self.data_dir.get())
        if directory:
            self.data_dir.set(directory)
            
    def browse_output_dir(self):
        directory = filedialog.askdirectory(initialdir=self.output_dir.get())
        if directory:
            self.output_dir.set(directory)
            
    def redirect_output(self):
        class StdoutRedirector:
            def __init__(self, text_widget):
                self.text_widget = text_widget
            def write(self, string):
                self.text_widget.insert(tk.END, string)
                self.text_widget.see(tk.END)
            def flush(self):
                pass
        sys.stdout = StdoutRedirector(self.console)
        
    def run_analysis(self):
        if self.running:
            return
                
        self.console.delete(1.0, tk.END)
        self.progress.start()
        self.running = True
        
        # Prepare arguments
        args_dict = {
            # Directory and basic options
            'data_dir': self.data_dir.get(),
            'output_dir': self.output_dir.get(),
            'data_format': self.data_format.get(),
            'eeglab_montage_units': self.eeglab_units.get(),
            'plot_preproc': self.plot_preproc.get(),
            'clean_muscle_artifacts': self.clean_muscle.get(),
            'show_evoked': self.show_evoked.get(),
            'research': self.research_stats.get(),
            'validate_teps': self.validate_teps.get(),  # Add validation option
            'save_validation': self.save_validation.get(),
            #'preproc_qc': self.preproc_qc.get(),
            'apply_ssp': self.apply_ssp.get(), 
            'apply_csd': self.apply_csd.get(),
            'skip_second_artifact_removal': self.skip_second_artifact_removal.get(),
            'stim_channel': 'STI 014',
            
            # Additional arguments with default values
            'substitute_zero_events_with': 10,
            'ica_method': 'fastica',
            'second_ica_method': 'infomax',
            'interpolation_method': 'cubic',  # Force cubic interpolation
            'interp_window': 1.0, 
            'no_second_ICA': False,
            'embed': False,
            'fix_artifact_window_start': -0.005,
            'fix_artifact_window_end': 0.015,
            'threshold_factor': 1.0,
            'muscle_window_start': 0.005,
            'muscle_window_end': 0.030,
            'lambda2': 1e-3,
            'stiffness': 4,
            'response_start': 0,        
            'response_end': 299,   
        }
        
        for param, var in self.params.items():
            try:
                if param in self.validation_rules:
                    if self.validation_rules[param]['type'] == 'str':
                        args_dict[param] = var.get()
                    else:
                        value = float(var.get())
                        if self.validation_rules[param]['type'] == int:
                            value = round(value)
                        args_dict[param] = value
            except ValueError:
                messagebox.showerror("Error", f"Invalid value for {param}")
                self.progress.stop()
                self.running = False
                return
        
        # Create Args object
        class Args:
            pass
        args = Args()
        for key, value in args_dict.items():
            setattr(args, key, value)
        
        # Run analysis in separate thread
        thread = threading.Thread(target=self.run_analysis_thread, args=(args,))
        thread.daemon = True
        thread.start()
        
    def run_analysis_thread(self, args):
        """Modified thread function with enhanced error capture and stop handling"""
        try:
            self.redirect_output()
            import matplotlib
            matplotlib.use('Agg')
            
            from ..run import process_subjects
            
            # Reset stop flag before starting
            import builtins
            setattr(builtins, 'STOP_PROCESSING', False)
            
            pcists = process_subjects(args)
            
            # Only show completion if we weren't stopped
            if not getattr(builtins, 'STOP_PROCESSING', False):
                self.root.after(0, self.analysis_complete, pcists)
        except Exception:
            import traceback
            error_msg = traceback.format_exc()
            self.root.after(0, self.analysis_error, error_msg)
        finally:
            # Always cleanup
            self.cleanup_after_stop()
            
    def analysis_complete(self, pcists):
        self.progress.stop()
        self.running = False
        
        # Build completion message
        message = f"Analysis completed successfully!\nPCIst values: {pcists}"
        
        if self.save_validation.get():
            output_dir = self.output_dir.get()
            message += f"\n\nValidation reports and plots saved to:\n{output_dir}"
            
            # Optional: List saved files
            try:
                validation_files = [f for f in os.listdir(output_dir) 
                                if f.startswith(('tep_validation_', 'pcist_'))]
                if validation_files:
                    message += "\n\nSaved files:"
                    for f in validation_files[:5]:  # Show first 5 files
                        message += f"\n- {f}"
                    if len(validation_files) > 5:
                        message += f"\n... and {len(validation_files)-5} more files"
            except Exception:
                pass
                
        messagebox.showinfo("Complete", message)
        
    def analysis_error(self, error_msg):
        """Enhanced error handling with full traceback"""
        import traceback
        import sys
        
        self.progress.stop()
        self.running = False
        
        # Get the full traceback
        exc_info = sys.exc_info()
        
        # Format error message with traceback
        error_detail = "Error Traceback:\n"
        
        if exc_info[0] is not None:  # If we have exception info
            exc_type, exc_value, exc_traceback = exc_info
            tb_list = traceback.extract_tb(exc_traceback)
            
            for filename, line, func, text in tb_list:
                error_detail += f"  File '{filename}', line {line}, in {func}\n"
                if text:
                    error_detail += f"    {text}\n"
            error_detail += f"\nError Type: {exc_type.__name__}\n"
        
        # Always include the error message
        error_detail += f"Error Message: {str(error_msg)}"
        
        # Show error in GUI
        messagebox.showerror("Error", error_detail)
        
        # Also print to console for debugging
        print("\nFull Error Details:")
        print(error_detail)
        
    def cleanup_after_stop(self):
        """Clean up after stopping the analysis"""
        import matplotlib.pyplot as plt
        plt.close('all')  # Close all matplotlib figures
        
        # Reset the stop flag
        import builtins
        setattr(builtins, 'STOP_PROCESSING', False)
        
        # Clear any remaining progress indication
        self.progress.stop()
        self.running = False
        print("\nAnalysis stopped and cleaned up")

    def stop_analysis(self):
        if self.running:
            print("\nStopping analysis (this may take a moment)...")
            self.running = False
            import builtins
            setattr(builtins, 'STOP_PROCESSING', True)
            self.root.after(1000, self.cleanup_after_stop)  # Cleanup after 1 second

if __name__ == "__main__":
    root = tk.Tk()
    app = TMSEEG_GUI(root)
    root.mainloop()